from __future__ import annotations

import logging
import pathlib
from collections import Counter, namedtuple
from operator import attrgetter

from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.class2D")

Class2DParticleClass = namedtuple(
    "Class2DParticleClass",
    [
        "particle_sum",
        "reference_image",
        "class_distribution",
        "accuracy_rotations",
        "accuracy_translations_angst",
        "estimated_resolution",
        "overall_fourier_completeness",
        "job",
    ],
)

Class2DParticleClass.__doc__ = "2D Classification stage."
Class2DParticleClass.particle_sum.__doc__ = "Sum of all particles in the class. Gives a tuple with the class number first, then the particle sum."
Class2DParticleClass.reference_image.__doc__ = "Reference image."
Class2DParticleClass.class_distribution.__doc__ = (
    "Class Distribution. Proportional to the number of particles per class."
)
Class2DParticleClass.accuracy_rotations.__doc__ = "Accuracy rotations."
Class2DParticleClass.accuracy_translations_angst.__doc__ = (
    "Accuracy translations angst."
)
Class2DParticleClass.estimated_resolution.__doc__ = "Estimated resolution."
Class2DParticleClass.overall_fourier_completeness.__doc__ = (
    "Overall Fourier completeness."
)
Class2DParticleClass.job.__doc__ = "Job number of the Class2D job."


class Class2D(JobType):
    def __eq__(self, other):
        if isinstance(other, Class2D):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.Class2D", self._basepath))

    def __repr__(self):
        return f"Class2D({repr(str(self._basepath))})"

    def __str__(self):
        return f"<Class2D parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):

        try:
            dfile, mfile = self._final_data_and_model(jobdir)
        except ValueError as e:
            logger.debug(
                f"The exception {e} was caught while trying to get data and model files. Returning an empty list",
                exc_info=True,
            )
            return []

        try:
            sdfile = self._read_star_file(jobdir, dfile)
            smfile = self._read_star_file(jobdir, mfile)
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            logger.debug(
                "gemmi could not open file while trying to get data and model files. Returning an empty list",
                exc_info=True,
            )
            return []

        info_table = self._find_table_from_column_name("_rlnClassDistribution", smfile)
        if info_table is None:
            logger.debug(f"_rlnClassDistribution not found in file {mfile}")
            return []

        class_distribution = self.parse_star_file(
            "_rlnClassDistribution", smfile, info_table
        )
        accuracy_rotations = self.parse_star_file(
            "_rlnAccuracyRotations", smfile, info_table
        )
        accuracy_translations_angst = self.parse_star_file(
            "_rlnAccuracyTranslationsAngst", smfile, info_table
        )
        estimated_resolution = self.parse_star_file(
            "_rlnEstimatedResolution", smfile, info_table
        )
        overall_fourier_completeness = self.parse_star_file(
            "_rlnOverallFourierCompleteness", smfile, info_table
        )
        reference_image = self.parse_star_file("_rlnReferenceImage", smfile, info_table)

        class_numbers = self.parse_star_file("_rlnClassNumber", sdfile, info_table)
        particle_sum = self._sum_all_particles(class_numbers)
        int_particle_sum = [(int(name), value) for name, value in particle_sum.items()]
        # something probably went wrong with file reading if this is the case
        # return empty list and hope to recover later
        if not int_particle_sum:
            return []
        try:
            checked_particle_list = self._class_checker(
                sorted(int_particle_sum), len(reference_image)
            )
        except IndexError:
            logger.debug(
                f"IndexErorr encountered in _class_checker for {jobdir}", exc_info=True
            )
            return []

        if len(reference_image) != len(checked_particle_list):
            logger.debug(
                f"Number of reference images did not match number of classes for {jobdir}"
            )

        particle_class_list = []

        try:
            for j in range(len(reference_image)):
                particle_class_list.append(
                    Class2DParticleClass(
                        checked_particle_list[j],
                        str(self._basepath.parent / reference_image[j].split("@")[1]),
                        float(class_distribution[j]),
                        accuracy_rotations[j],
                        accuracy_translations_angst[j],
                        estimated_resolution[j],
                        overall_fourier_completeness[j],
                        jobdir,
                    )
                )
        except IndexError:
            logger.debug(
                "An IndexError was encountered while collecting 2D classification data: there was possibly a mismatch between data from different files"
            )
        return particle_class_list

    def _final_data_and_model(self, job_path):
        number_list = [
            entry.stem[6:9]
            for entry in (self._basepath / job_path).glob("run_it*.star")
        ]
        last_iteration_number = max(
            (int(n) for n in number_list if n.isnumeric()), default=0
        )
        if not last_iteration_number:
            raise ValueError(f"No result files found in {job_path}")
        data_file = f"run_it{last_iteration_number:03d}_data.star"
        model_file = f"run_it{last_iteration_number:03d}_model.star"
        for check_file in (
            self._basepath / job_path / data_file,
            self._basepath / job_path / model_file,
        ):
            if not check_file.exists():
                raise ValueError(f"File {check_file} missing from job directory")
        return data_file, model_file

    def _count_all(self, list):
        count = Counter(list)
        return count

    def _sum_all_particles(self, list):
        counted = self._count_all(
            list
        )  # This sorts them into classes and the number of associated particles before summing them all
        return counted
        # return sum(counted.values())

    def _class_checker(
        self, tuple_list, length
    ):  # Makes sure every class has a number of associated particles
        if not tuple_list:
            raise IndexError
        for i in range(1, length + 1):
            try:
                if i not in tuple_list[i - 1]:
                    tuple_list.insert(i - 1, (i, 0))
                    # print("No values found for class", i)
            except IndexError:
                tuple_list.insert(i - 1, (i, 0))
                # print("No values found for class", i)
        return tuple_list

    def top_twenty(self, dictionary):
        return_dict = {}
        for item in dictionary:
            temp_list = sorted(dictionary[item], key=attrgetter("class_distribution"))[
                -20:
            ]
            temp_list.reverse()
            return_dict[item] = temp_list
        return return_dict

    @staticmethod
    def db_unpack(particle_class):
        def _get_img_mod_time(img):
            if not img:
                return None
            imgpath = pathlib.Path(img)
            if not imgpath.exists():
                return None
            return imgpath.stat().st_mtime

        res = [
            {
                "type": "2D",
                "class_number": cl.particle_sum[0],
                "particles_per_class": cl.particle_sum[1],
                "rotation_accuracy": cl.accuracy_rotations,
                "translation_accuracy": cl.accuracy_translations_angst,
                "estimated_resolution": cl.estimated_resolution,
                "overall_fourier_completeness": cl.overall_fourier_completeness,
                "class_image_full_path": cl.reference_image.replace(
                    ".mrcs", f"_{cl.particle_sum[0]}.jpeg"
                ),
                "job_string": cl.job,
                "class_images_stack": cl.reference_image,
                "class_images_modification_time": _get_img_mod_time(cl.reference_image),
                "class_distribution": cl.class_distribution,
            }
            for cl in particle_class
        ]
        return res



from __future__ import annotations

import logging
from collections import namedtuple

from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.motioncorrection")

MCMicrograph = namedtuple(
    "MCMicrograph",
    [
        "micrograph_name",
        "micrograph_snapshot_full_path",
        "micrograph_number",
        "total_motion",
        "early_motion",
        "late_motion",
        "micrograph_timestamp",
        "drift_data",
    ],
)

MCMicrograph.__doc__ = "Motion Correction stage."
MCMicrograph.micrograph_name.__doc__ = "Micrograph name. Useful for reference."
MCMicrograph.micrograph_snapshot_full_path.__doc__ = (
    "Path to jpeg of the motion corrected micrograph."
)
MCMicrograph.micrograph_number.__doc__ = "Micrograph number: sequential in time."
MCMicrograph.total_motion.__doc__ = (
    "Total motion. The amount the sample moved during exposure. Units angstrom (A)."
)
MCMicrograph.early_motion.__doc__ = "Early motion."
MCMicrograph.late_motion.__doc__ = "Late motion."
MCMicrograph.micrograph_timestamp.__doc__ = (
    "Time stamp at which the micrograph was created."
)


MCMicrographDrift = namedtuple(
    "MCMicrographDrift",
    [
        "frame",
        "deltaX",
        "deltaY",
    ],
)

MCDriftCacheRecord = namedtuple(
    "MCDriftCacheRecord",
    [
        "data",
        "file_size",
        "movie_name",
    ],
)


class MotionCorr(JobType):
    def __init__(self, path, drift_cache=None):
        super().__init__(path)
        if drift_cache is None:
            self._drift_cache = {}
        else:
            self._drift_cache = drift_cache

    def __eq__(self, other):
        if isinstance(other, MotionCorr):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.MotionCorr", self._basepath))

    def __repr__(self):
        return f"MotionCorr({repr(str(self._basepath))})"

    def __str__(self):
        return f"<MotionCorr parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):
        try:
            file = self._read_star_file(jobdir, "corrected_micrographs.star")
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            return []

        info_table = self._find_table_from_column_name("_rlnAccumMotionTotal", file)
        if info_table is None:
            logger.debug(f"_rlnAccumMotionTotal not found in file {file}")
            return []

        accum_motion_total = self.parse_star_file(
            "_rlnAccumMotionTotal", file, info_table
        )
        accum_motion_late = self.parse_star_file(
            "_rlnAccumMotionLate", file, info_table
        )
        accum_motion_early = self.parse_star_file(
            "_rlnAccumMotionEarly", file, info_table
        )
        micrograph_name = self.parse_star_file("_rlnMicrographName", file, info_table)

        micrograph_list = []
        for j in range(len(micrograph_name)):
            drift_data, movie_name = self.collect_drift_data(micrograph_name[j], jobdir)
            if movie_name:
                try:
                    movie_creation_time = (
                        (self._basepath.parent / movie_name).resolve().stat().st_ctime
                    )
                except FileNotFoundError:
                    logger.debug(
                        f"failed to find movie {self._basepath.parent / movie_name} so using default timestamp"
                    )
                    movie_creation_time = None
            else:
                movie_creation_time = None
            micrograph_list.append(
                MCMicrograph(
                    micrograph_name[j],
                    str(self._basepath.parent / micrograph_name[j]).replace(
                        ".mrc", ".jpeg"
                    ),
                    j + 1,
                    accum_motion_total[j],
                    accum_motion_early[j],
                    accum_motion_late[j],
                    movie_creation_time,
                    drift_data,
                )
            )
        return micrograph_list

    def collect_drift_data(self, mic_name, jobdir):
        drift_data = []
        drift_star_file_path = mic_name.split(jobdir + "/")[-1].replace("mrc", "star")
        if self._drift_cache.get(jobdir):
            if self._drift_cache[jobdir].get(mic_name):
                try:
                    if (
                        self._drift_cache[jobdir][mic_name].file_size
                        == (self._basepath / jobdir / drift_star_file_path)
                        .stat()
                        .st_size
                    ):
                        return (
                            self._drift_cache[jobdir][mic_name].data,
                            self._drift_cache[jobdir][mic_name].movie_name,
                        )
                except FileNotFoundError:
                    logger.debug(
                        "Could not find expected file containing drift data",
                        exc_info=True,
                    )
                    return [], ""
        else:
            self._drift_cache[jobdir] = {}
        try:
            drift_star_file = self._read_star_file(jobdir, drift_star_file_path)
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            return drift_data, ""
        try:
            info_table = self._find_table_from_column_name(
                "_rlnMicrographFrameNumber", drift_star_file
            )
            movie_table = 0
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            return drift_data, ""
        if info_table is None:
            logger.debug(
                f"_rlnMicrographFrameNumber or _rlnMicrographMovieName not found in file {drift_star_file}"
            )
            return drift_data, ""
        frame_numbers = self.parse_star_file(
            "_rlnMicrographFrameNumber", drift_star_file, info_table
        )
        deltaxs = self.parse_star_file(
            "_rlnMicrographShiftX", drift_star_file, info_table
        )
        deltays = self.parse_star_file(
            "_rlnMicrographShiftY", drift_star_file, info_table
        )
        movie_name = self.parse_star_file_pair(
            "_rlnMicrographMovieName", drift_star_file, movie_table
        )
        for f, dx, dy in zip(frame_numbers, deltaxs, deltays):
            drift_data.append(MCMicrographDrift(int(f), float(dx), float(dy)))
        try:
            self._drift_cache[jobdir][mic_name] = MCDriftCacheRecord(
                drift_data,
                (self._basepath / jobdir / drift_star_file_path).stat().st_size,
                movie_name,
            )
        except FileNotFoundError:
            return [], ""
        return drift_data, movie_name

    @staticmethod
    def for_cache(mcmicrograph):
        return str(mcmicrograph.micrograph_name)

    @staticmethod
    def for_validation(mcmicrograph):
        return {str(mcmicrograph.micrograph_name): mcmicrograph.micrograph_number}

    # this allows an MCMicrograph object to be copied but with some attributes changed
    @staticmethod
    def mutate_result(mcmicrograph, **kwargs):
        attr_names_list = MCMicrograph._fields
        attr_list = [
            kwargs.get(name, getattr(mcmicrograph, name)) for name in attr_names_list
        ]
        return MCMicrograph(*attr_list)

    @staticmethod
    def db_unpack(micrograph_list):
        res = [
            {
                "micrograph_full_path": micrograph.micrograph_name,
                "total_motion": micrograph.total_motion,
                "early_motion": micrograph.early_motion,
                "late_motion": micrograph.late_motion,
                "average_motion_per_frame": (
                    float(micrograph.total_motion) / len(micrograph.drift_data)
                ),
                "image_number": micrograph.micrograph_number,
                "micrograph_snapshot_full_path": micrograph.micrograph_snapshot_full_path,
                "drift_data": micrograph.drift_data,
                "created_time_stamp": micrograph.micrograph_timestamp,
            }
            for micrograph in micrograph_list
        ]
        return res



"""
Relion Python API
https://github.com/DiamondLightSource/python-relion
"""

from __future__ import annotations

import functools
import os
import pathlib
import time
from collections import namedtuple

from gemmi import cif

from relion._parser.autopick import AutoPick
from relion._parser.class2D import Class2D
from relion._parser.class3D import Class3D
from relion._parser.cryolo import Cryolo, CryoloAutoPick
from relion._parser.ctffind import CTFFind
from relion._parser.initialmodel import InitialModel
from relion._parser.motioncorrection import MotionCorr
from relion._parser.relativeicethickness import RelativeIceThickness
from relion._parser.relion_pipeline import RelionPipeline

try:
    from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions
except ModuleNotFoundError:
    pass
import logging

from relion.dbmodel import DBGraph, DBModel, DBNode
from relion.node.graph import Graph

logger = logging.getLogger("relion.Project")

__all__ = []
__author__ = "Diamond Light Source - Scientific Software"
__email__ = "scientificsoftware@diamond.ac.uk"
__version__ = "0.11.3"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))

pipeline_lock = ".relion_lock"


RelionJobResult = namedtuple(
    "RelionJobResult",
    [
        "stage_object",
        "job_name",
        "end_time_stamp",
    ],
)

RelionJobInfo = namedtuple(
    "RelionJobInfo",
    [
        "job_name",
        "end_time_stamp",
    ],
)


class Project(RelionPipeline):
    """
    Reads information from a Relion project directory and makes it available in
    a structured, object-oriented, and pythonic fashion.
    """

    def __init__(
        self,
        path,
        database="ISPyB",
        run_options=None,
        message_constructors=None,
        cluster=False,
        version: int = 3,
    ):
        """
        Create an object representing a Relion project.
        :param path: A string or file system path object pointing to the root
                     directory of an existing Relion project.
        """
        self.basepath = pathlib.Path(path)
        self._version = version
        super().__init__(
            "Import/job001", locklist=[self.basepath / "default_pipeline.star"]
        )
        if message_constructors is not None:
            self.construct_messages = message_constructors
        else:
            self.construct_messages = {}
        if not self.basepath.is_dir():
            raise ValueError(f"path {self.basepath} is not a directory")
        self._data_pipeline = Graph("DataPipeline", [])
        self._db_model = DBModel(database)
        self._drift_cache = {}
        if run_options is None:
            self.run_options = RelionItOptions()
        else:
            self.run_options = run_options
        try:
            self.load(cluster=cluster)
        except (FileNotFoundError, OSError, RuntimeError):
            pass
            # raise RuntimeWarning(
            #    f"Relion Project was unable to load the relion pipeline from {self.basepath}/default_pipeline.star"
            # )
        # self.res = RelionResults()

    @property
    def _plock(self):
        return PipelineLock(self.basepath / pipeline_lock)

    def __eq__(self, other):
        if isinstance(other, Project):
            return self.basepath == other.basepath
        return False

    def __hash__(self):
        return hash(("relion.Project", self.basepath))

    def __repr__(self):
        return f"relion.Project({repr(str(self.basepath))})"

    def __str__(self):
        return f"<relion.Project at {self.basepath}>"

    @property
    def _results_dict(self):
        resd = {
            "CtfFind": self.ctffind,
            "MotionCorr": self.motioncorrection,
            "AutoPick": self.autopick_cryolo
            if self.run_options.autopick_do_cryolo and self._version == 4
            else self.autopick,
            "External/crYOLO_AutoPick/": self.cryolo,
            "Class2D": self.class2D,
            "InitialModel": self.initialmodel,
            "Class3D": self.class3D,
            "External/Icebreaker_5fig/": self.relativeicethickness,
            "IceBreaker/Icebreaker_5fig/": self.relativeicethickness_ib,
        }
        return resd

    @property
    @functools.lru_cache(maxsize=1)
    def ctffind(self):
        """access the CTFFind stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of CTFMicrograph namedtuples as values."""
        return CTFFind(self.basepath / "CtfFind")

    @property
    @functools.lru_cache(maxsize=1)
    def motioncorrection(self):
        """access the motion correction stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of MCMicrograph namedtuples as values."""
        return MotionCorr(self.basepath / "MotionCorr", self._drift_cache)

    @property
    @functools.lru_cache(maxsize=1)
    def autopick(self):
        return AutoPick(self.basepath / "AutoPick")

    @property
    @functools.lru_cache(maxsize=1)
    def autopick_cryolo(self):
        return CryoloAutoPick(self.basepath / "AutoPick")

    @property
    @functools.lru_cache(maxsize=1)
    def cryolo(self):
        return Cryolo(self.basepath / "External")

    @property
    @functools.lru_cache(maxsize=1)
    def class2D(self):
        """access the 2D classification stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of Class2DParticleClass namedtuples as values."""
        return Class2D(self.basepath / "Class2D")

    @property
    @functools.lru_cache(maxsize=1)
    def initialmodel(self):
        return InitialModel(self.basepath / "InitialModel")

    @property
    @functools.lru_cache(maxsize=1)
    def class3D(self):
        """access the 3D classification stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of Class3DParticleClass namedtuples as values."""
        return Class3D(self.basepath / "Class3D")

    @property
    @functools.lru_cache(maxsize=1)
    def relativeicethickness(self):
        return RelativeIceThickness(self.basepath / "External")

    @property
    @functools.lru_cache(maxsize=1)
    def relativeicethickness_ib(self):
        return RelativeIceThickness(self.basepath / "IceBreaker")

    def origin_present(self):
        try:
            self.load_nodes_from_star(self.basepath / "default_pipeline.star")
        except (TypeError, FileNotFoundError, OSError, RuntimeError):
            return False
        if len(self._nodes) == 0:
            return False
        return (self.basepath / self.origin / "RELION_JOB_EXIT_SUCCESS").is_file()

    def load(self, clear_cache=True, cluster=False):
        if clear_cache:
            self._clear_caches()
        self._data_pipeline = Graph("DataPipeline", [])
        # reset the in and out lists of database nodes
        # have to avoid removing the permanent connections from other database nodes
        for dbn in self._db_model.values():
            dbn._in = [
                i_node
                for i_node in dbn._in
                if isinstance(i_node, DBNode) or isinstance(i_node, DBGraph)
            ]
        self._jobs_collapsed = False
        try:
            self.load_nodes_from_star(self.basepath / "default_pipeline.star")
        except FileNotFoundError:
            return
        self.check_job_node_statuses(self.basepath)
        self.collect_job_times(
            list(self.schedule_files), self.basepath / "pipeline_PREPROCESS.log"
        )
        if cluster:
            self.collect_cluster_info(self.basepath)
        for jobnode in self:
            if (
                self._results_dict.get(jobnode.name)
                or jobnode.environment.get("alias") in self._results_dict
            ):
                if jobnode.name == "InitialModel":
                    self._update_pipeline(
                        jobnode,
                        jobnode.name,
                        prop=("ini_model_job_string", "ini_model_job_string"),
                        in_db_model=False,
                    )
                elif jobnode.name == "AutoPick":
                    self._update_pipeline(
                        jobnode, jobnode.name, prop=("job_string", "parpick_job_string")
                    )
                elif "crYOLO" in jobnode.environment.get("alias"):
                    self._update_pipeline(
                        jobnode,
                        jobnode.environment.get("alias"),
                        prop=("job_string", "parpick_job_string"),
                    )
                elif jobnode.name == "External":
                    self._update_pipeline(
                        jobnode,
                        jobnode.environment.get("alias"),
                    )
                elif jobnode.name == "IceBreaker":
                    self._update_pipeline(
                        jobnode,
                        jobnode.environment.get("alias"),
                    )
                else:
                    self._update_pipeline(jobnode, jobnode.name)
            else:
                self._data_pipeline.add_node(jobnode)
                if jobnode.name == "Import":
                    self._data_pipeline.origins = [jobnode]

    def _update_pipeline(self, jobnode, label, prop=None, in_db_model=True):
        if jobnode.environment["status"]:
            jobnode.environment["result"] = self._results_dict[label]
        else:
            jobnode.environment["result"] = {}
        if in_db_model:
            jobnode.environment["extra_options"] = self.run_options
            self._db_model[label].environment["extra_options"] = self.run_options
            self._db_model[label].environment[
                "message_constructors"
            ] = self.construct_messages
        if prop is not None:
            jobnode.propagate(prop)
        jobnode.link_to(
            self._db_model[label],
            result_as_traffic=True,
            share=[("end_time_stamp", "end_time")],
        )
        self._data_pipeline.add_node(jobnode)
        if in_db_model:
            self._data_pipeline.add_node(self._db_model[label])

    def show_job_nodes(self):
        self.load()
        super().show_job_nodes(self.basepath)

    @property
    def schedule_files(self):
        return self.basepath.glob("pipeline*.log")

    @property
    def messages(self):
        self._clear_caches()
        msgs = []
        results = self._data_pipeline()
        if results is None:
            return msgs
        for node in self._db_model.db_nodes:
            try:
                if results[node.nodeid] is not None:
                    d = {}
                    if isinstance(results[node.nodeid], dict):
                        for key, val in results[node.nodeid].items():
                            try:
                                d[key].append(val)
                            except KeyError:
                                d[key] = val
                    else:
                        for p in results[node.nodeid]:
                            for key, val in p.items():
                                try:
                                    d[key].extend(val)
                                except KeyError:
                                    d[key] = val
                    msgs.append(d)
            except KeyError:
                logger.debug(
                    f"No results found for {node.name}: probably the job has not completed yet"
                )
        return msgs

    @property
    def current_jobs(self):
        self.load(cluster=True)
        currj = super().current_jobs
        if currj is None:
            return None
        else:
            for n in currj:
                n.change_name(self.basepath / n.name)
            return currj

    @staticmethod
    def _clear_caches():
        Project.motioncorrection.fget.cache_clear()
        Project.ctffind.fget.cache_clear()
        Project.autopick.fget.cache_clear()
        Project.cryolo.fget.cache_clear()
        Project.autopick_cryolo.fget.cache_clear()
        Project.class2D.fget.cache_clear()
        Project.initialmodel.fget.cache_clear()
        Project.class3D.fget.cache_clear()
        Project.relativeicethickness.fget.cache_clear()
        Project.relativeicethickness_ib.fget.cache_clear()

    def get_imported(self):
        try:
            import_job_path = self.basepath / self.origin
            gemmi_readable_path = os.fspath(import_job_path / "movies.star")
            star_doc = cif.read_file(gemmi_readable_path)
            for index, block in enumerate(star_doc):
                if list(block.find_loop("_rlnMicrographMovieName")):
                    block_index = index
                    break
            else:
                return []
            data_block = star_doc[block_index]
            values = list(data_block.find_loop("_rlnMicrographMovieName"))
            if not values:
                return []
            return values
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            return []


# helper class for dealing with the default_pipeline.star lock
class PipelineLock:
    def __init__(self, lockdir):
        self.lockdir = lockdir
        self.fail_count = 0
        self.obtained = False

    def __enter__(self):
        while self.fail_count < 20:
            try:
                self.lockdir.mkdir()
                self.obtained = True
                break
            except FileExistsError:
                time.sleep(0.1)
                self.fail_count += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.obtained:
            try_count = 0
            while try_count < 5:
                try:
                    self.lockdir.rmdir()
                    break
                except OSError:
                    try_count += 1
                    time.sleep(1)
        self.obtained = False



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict

from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions


def _from_import(
    tracker: Dict[str, Path], *args, in_key: str = "input_star_mics"
) -> Dict[str, Any]:
    return {in_key: str(tracker["relion.import.movies"] / "movies.star")}


def _from_motioncorr(
    tracker: Dict[str, Path], options: RelionItOptions, in_key: str = "input_star_mics"
) -> Dict[str, Any]:
    if options.images_are_movies:
        if options.motioncor_do_own:
            return {
                in_key: str(
                    tracker["relion.motioncorr.own"] / "corrected_micrographs.star"
                )
            }
        return {
            in_key: str(
                tracker["relion.motioncorr.motioncor2"] / "corrected_micrographs.star"
            )
        }
    return _from_import(tracker, in_key=in_key)


def _from_ctf(
    tracker: Dict[str, Path],
    options: RelionItOptions,
    in_key: str = "fn_input_autopick",
) -> Dict[str, Any]:
    if options.use_ctffind_instead:
        return {
            in_key: str(tracker["relion.ctffind.ctffind4"] / "micrographs_ctf.star")
        }
    return {in_key: str(tracker["relion.ctffind.gctf"] / "micrographs_ctf.star")}


def _from_ib(
    tracker: Dict[str, Path], options: RelionItOptions, in_key: str = "in_mics"
) -> Dict[str, Any]:
    return {
        in_key: str(
            tracker["icebreaker.micrograph_analysis.micrographs"]
            / "grouped_micrographs.star"
        )
    }


def _extract(
    tracker: Dict[str, Path], options: RelionItOptions, ref: bool = False
) -> Dict[str, Any]:
    res = {}
    if options.autopick_do_cryolo:
        res["coords_suffix"] = str(tracker["cryolo.autopick"] / "autopick.star")
    elif ref:
        res["coords_suffix"] = str(tracker["relion.autopick.ref3d"] / "autopick.star")
    else:
        res["coords_suffix"] = str(tracker["relion.autopick.log"] / "autopick.star")
    res.update(_from_ctf(tracker, options, in_key="star_mics"))
    return res


def _select(
    tracker: Dict[str, Path], options: RelionItOptions, ref: str = ""
) -> Dict[str, Any]:
    return {"fn_data": str(tracker["relion.extract" + ref] / "particles.star")}


_extra_options: Dict[str, Callable] = {
    "relion.motioncorr.motioncor2": _from_import,
    "relion.motioncorr.own": _from_import,
    "icebreaker.micrograph_analysis.micrographs": partial(
        _from_motioncorr, in_key="in_mics"
    ),
    "icebreaker.micrograph_analysis.enhancecontrast": partial(
        _from_motioncorr, in_key="in_mics"
    ),
    "icebreaker.micrograph_analysis.summary": partial(
        _from_ib,
    ),
    "relion.ctffind.ctffind4": _from_motioncorr,
    "relion.ctffind.gctf": _from_motioncorr,
    "relion.autopick.log": _from_ctf,
    "relion.autopick.ref3d": _from_ctf,
    "cryolo.autopick": partial(_from_ctf, in_key="input_file"),
    "relion.extract": _extract,
    "relion.select.split": _select,
}


def generate_extra_options(
    job: str, tracker: Dict[str, Path], options: RelionItOptions
) -> Dict[str, Any]:
    if job.startswith("relion.extract"):
        refb = bool(job.replace("relion.extract", ""))
        return _extra_options["relion.extract"](tracker, options, ref=refb)
    elif job.startswith("relion.select.split"):
        ref = job.replace("relion.select.split", "")
        return _extra_options["relion.select.split"](tracker, options, ref=ref)
    else:
        try:
            return _extra_options[job](tracker, options)
        except KeyError:
            return {}



from __future__ import annotations

import logging
from collections import namedtuple

from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.initalmodel")

InitialModelInfo = namedtuple(
    "InitialModelInfo",
    [
        "number_of_particles",
    ],
)


class InitialModel(JobType):
    def __eq__(self, other):
        if isinstance(other, InitialModel):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.InitialModel", self._basepath))

    def __repr__(self):
        return f"InitialModel({repr(str(self._basepath))})"

    def __str__(self):
        return f"<InitialModel parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):
        try:
            model_info_name = self._final_data(jobdir)
        except (RuntimeError, FileNotFoundError, OSError):
            return []

        try:
            model_info_file = self._read_star_file(jobdir, model_info_name)
        except (RuntimeError, FileNotFoundError, OSError):
            return []

        info_table = self._find_table_from_column_name(
            "_rlnClassNumber", model_info_file
        )
        if info_table is None:
            logger.debug(f"_rlnClassNumber not found in file {model_info_file}")
            return []

        num_particles = {}
        class_numbers = self.parse_star_file(
            "_rlnClassNumber", model_info_file, info_table
        )
        for n in class_numbers:
            num_particles[int(n)] = num_particles.get(int(n), 0) + 1

        return [InitialModelInfo(num_particles)]

    def _final_data(self, job_path):
        number_list = [
            entry.stem[6:9]
            for entry in (self._basepath / job_path).glob("run_it*.star")
        ]
        last_iteration_number = max(
            (int(n) for n in number_list if n.isnumeric()), default=0
        )
        if not last_iteration_number:
            raise ValueError(f"No result files found in {job_path}")
        data_file = f"run_it{last_iteration_number:03d}_data.star"
        for check_file in (self._basepath / job_path / data_file,):
            if not check_file.exists():
                raise ValueError(f"File {check_file} missing from job directory")
        return data_file

    @staticmethod
    def for_cache(initmodelinfo):
        return initmodelinfo.number_of_particles

    @staticmethod
    def db_unpack(initmodelinfo):
        res = {
            "init_model_number_of_particles": initmodelinfo[0].number_of_particles,
        }

        return res



#!/usr/bin/env python
"""
External job for calling cryolo fine tune within Relion 3.1
in_parts is from a subset selection job.
cryolo_fine_tune_job.py --o 'External/crYOLO_FineTune' --in_parts 'Select/job005/particles.star' --box_size 300
"""

from __future__ import annotations

import argparse
import json
import os
import os.path
import shutil
import subprocess
import time

import gemmi

RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"

CRYOLO_FINETUNE_JOB_DIR = "External/crYOLO_FineTune"


def run_job(project_dir, job_dir, args_list):
    parser = argparse.ArgumentParser()
    parser.add_argument("--in_parts", help="Input micrographs STAR file")
    parser.add_argument(
        "--j", dest="threads", help="Number of threads to run (ignored)"
    )
    parser.add_argument("--box_size", help="Size of box (~ particle size)")
    parser.add_argument("--gmodel", help="cryolo general model")
    parser.add_argument("--config", help="cryolo config")
    args = parser.parse_args(args_list)
    box_size = args.box_size
    weights = args.gmodel
    conf_file = args.config

    # Making a cryolo config file with the correct box size and model location
    with open(conf_file, "r") as json_file:
        data = json.load(json_file)
        data["model"]["anchors"] = [int(box_size), int(box_size)]
        data["train"]["pretrained_weights"] = weights
    with open("config.json", "w") as outfile:
        json.dump(data, outfile)

    # Reading particle star file from relion
    count = 0
    particle_file = os.path.join(project_dir, args.in_parts)
    while not os.path.exists(particle_file):
        count += 1
        if count > 60:
            print(
                f" cryolo_fine_tune_job: giving up after waiting for over {count * 10} seconds for particle file {particle_file} to appear"
            )
            raise AssertionError("Timeout waiting for input file")
        if count % 6 == 0:
            print(
                f" cryolo_fine_tune_job: still waiting for particle file {particle_file} to appear after {count * 10} seconds"
            )
        time.sleep(10)
    in_doc = gemmi.cif.read_file(particle_file)
    data_as_dict = json.loads(in_doc.as_json())["micrographs"]

    try:
        os.mkdir("train_annotation")
    except FileExistsError:
        shutil.rmtree("train_annotation")
        os.mkdir("train_annotation")
    try:
        os.mkdir("train_image")
    except FileExistsError:
        shutil.rmtree("train_image")
        os.mkdir("train_image")

    # Arranging files for cryolo to train from
    for micro in range(len(data_as_dict["_rlnmicrographname"])):
        try:
            os.link(
                os.path.join(project_dir, data_as_dict["_rlnmicrographname"][micro]),
                os.path.join(
                    project_dir,
                    job_dir,
                    "train_image",
                    os.path.split(data_as_dict["_rlnmicrographname"][micro])[-1],
                ),
            )
        except FileExistsError:
            pass

        box_name = (
            os.path.splitext(
                os.path.split(data_as_dict["_rlnmicrographname"][micro])[-1]
            )[0]
            + ".box"
        )

        individual_files = open(os.path.join("train_annotation", box_name), "a+")
        individual_files.write(
            f"{data_as_dict['_rlncoordinatex'][micro] - int(box_size)/2}\t"
        )
        individual_files.write(
            f"{data_as_dict['_rlncoordinatey'][micro] - int(box_size)/2}\t"
        )
        individual_files.write(f"{box_size}\t")
        individual_files.write(f"{box_size}\n")
        individual_files.close()

    # Running cryolo
    subprocess.run(
        [
            "cryolo_train.py",
            "--conf",
            "config.json",
            "--warmup",
            "0",
            "--gpu",
            "0",
            "--fine_tune",
        ]
    )

    # Writing a star file (This one is meaningless for now)
    with open("_manualpick.star", "w") as part_doc:
        part_doc.write(os.path.join(project_dir, args.in_parts))

    # Required star file
    out_doc = gemmi.cif.Document()
    output_nodes_block = out_doc.add_new_block("output_nodes")
    loop = output_nodes_block.init_loop(
        "", ["_rlnPipeLineNodeName", "_rlnPipeLineNodeType"]
    )
    loop.add_row([os.path.join(job_dir, "_manualpick.star"), "2"])
    out_doc.write_file("RELION_OUTPUT_NODES.star")
    print(" cryolo_fine_tune_job: crYOLO Finished Fine-Tuning")


def main():
    """Change to the job working directory, then call run_job()"""
    parser = argparse.ArgumentParser()
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    known_args, other_args = parser.parse_known_args()
    project_dir = os.getcwd()
    os.makedirs(known_args.out_dir, exist_ok=True)
    os.chdir(known_args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        print(" cryolo_fine_tune_job: Removing previous failure indicator file")
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        print(" cryolo_fine_tune_job: Removing previous success indicator file")
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        run_job(project_dir, known_args.out_dir, other_args)
    except Exception:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


def run_cryolo_job(job_dir, command_list, pipeline_opts, wait_for_completion=True):
    """Run a cryolo job (submitting to the queue if requested) and optionally wait for completion"""
    success_file = os.path.join(job_dir, RELION_JOB_SUCCESS_FILENAME)
    failure_file = os.path.join(job_dir, RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(failure_file):
        print(f" cryolo_pipeline: Removing previous job failure file {failure_file}")
        os.remove(failure_file)
    if os.path.isfile(success_file):
        print(f" cryolo_pipeline: Removing previous job success file {success_file}")
        os.remove(success_file)

    if pipeline_opts.cryolo_submit_to_queue:
        # TODO: remove -o and -e arguments after switch to Relion 3.1. This syntax is specific to
        # qsub but necessary for now to ensure stdout and stderr files go in the job directory
        # rather than the project directory. In Relion 3.1 the normal Relion qsub template can be
        # used which will put the files in the right place automatically.
        submit_command = [
            pipeline_opts.queue_submit_command,
            "-o",
            os.path.join(job_dir, "cryolo_job.out"),  # Temporary!
            "-e",
            os.path.join(job_dir, "cryolo_job.err"),  # Temporary!
            pipeline_opts.cryolo_queue_submission_template,
        ]
        submit_command.extend(command_list)
        print(
            " cryolo_pipeline: running cryolo command: {}".format(
                " ".join(submit_command)
            )
        )
        subprocess.Popen(submit_command)
    else:
        print(
            " cryolo_pipeline: running cryolo command: {}".format(
                " ".join(command_list)
            )
        )
        subprocess.Popen(command_list)

    if wait_for_completion:
        count = 0
        while not (os.path.isfile(failure_file) or os.path.isfile(success_file)):
            count += 1
            if count % 6 == 0:
                print(
                    f" cryolo_pipeline: Still waiting for cryolo job to finish after {count * 10} seconds"
                )
            time.sleep(10)


if __name__ == "__main__":
    main()



from __future__ import annotations

from datetime import datetime

from relion.dbmodel import modeltables
from relion.node import Node


class DBNode(Node):
    def __init__(self, name, tables, **kwargs):
        super().__init__(name, **kwargs)
        self.shape = "octagon"
        if not isinstance(tables, list):
            raise TypeError(f"{self} could not be initialised: tables must be a list")
        self.tables = tables
        self._append_sent = {}
        for table in self.tables:
            table._last_update[self.name] = 0
            self._append_sent[table] = {a: set() for a in table._append}

        self._sent = [[] for _ in self.tables]
        self._unsent = [[] for _ in self.tables]
        self._all_sent = [[] for _ in self.tables]

    def __eq__(self, other):
        if isinstance(other, DBNode):
            if self.name == other.name and len(self._out) == len(other._out):
                for n in self._out:
                    if n not in other._out:
                        return False
                return True
        return False

    def __hash__(self):
        return hash(("relion.dbmodel.dbnode.DBNode", self._name))

    def __repr__(self):
        return f"Node({repr(str(self._name))})"

    def __bool__(self):
        if self.tables:
            return True
        return False

    def func(self, *args, **kwargs):
        if self.environment.empty:
            return []
        extra_options = self.environment["extra_options"]
        if self.environment["end_time"] is not None:
            end_time = datetime.timestamp(self.environment["end_time"])
        else:
            end_time = None
        msg_cons = self.environment["message_constructors"]
        self.insert(end_time, extra_options)
        return self.message(msg_cons)

    def update_times(self, source=None):
        if source is None:
            all_times = []
            for tab in self.tables:
                for k in tab._last_update.keys():
                    all_times.append(tab._last_update[k])
        return [tab._last_update[source] for tab in self.tables]

    def insert(self, end_time, extra_options):
        source_option = self.environment["source"]
        for i, tab in enumerate(self.tables):
            self._do_check()
            if end_time is None:
                return
            if (
                tab._last_update[source_option or self.name] is None
                or end_time > tab._last_update[source_option or self.name]
            ):
                tab._last_update[source_option or self.name] = end_time

            pid = modeltables.insert(
                tab,
                end_time,
                source_option or self.name,
                extra_options,
                self.environment,
            )
            if pid is not None:
                self._unsent[i].append(pid)
                if pid in self._sent[i]:
                    self._sent[i].remove(pid)

    def _do_check(self):
        try:
            if self.environment["check_for"] is not None:
                table = self.environment["foreign_table"]
                check_for_foreign_name = self.environment["check_for_foreign_name"]
                if check_for_foreign_name is None:
                    check_for_foreign_name = self.environment["check_for"]
                indices = table.get_row_index(
                    check_for_foreign_name,
                    self.environment[self.environment["check_for"]],
                )
                if indices is None:
                    self.environment[self.environment["table_key"]] = None
                    return
                try:
                    entries = [table[table._primary_key][ci] for ci in indices]
                    if self.environment["first"]:
                        self.environment[self.environment["table_key"]] = entries[0]
                    else:
                        self.environment[self.environment["table_key"]] = entries
                    return
                except TypeError:
                    self.environment[self.environment["table_key"]] = table[
                        table._primary_key
                    ][indices]
                    return
            else:
                return
        except KeyError:
            return

    def message(self, constructors=None):
        if constructors is None:
            return {}
        messages = {msg_type: [] for msg_type in constructors.keys()}
        for tab_index, ids in enumerate(self._unsent):
            for pid in ids:
                for msg_type, constructor in constructors.items():
                    unsent_appended = {}
                    for acol in self.tables[tab_index]._append:
                        row = self.tables[tab_index].get_row_by_primary_key(pid)
                        try:
                            unsent_appended[acol] = [
                                e
                                for e in row[acol]
                                if e
                                not in self._append_sent[self.tables[tab_index]][acol]
                            ]
                        except TypeError:
                            if (
                                row[acol]
                                not in self._append_sent[self.tables[tab_index]][acol]
                            ):
                                unsent_appended[acol] = [row[acol]]
                        if unsent_appended.get(acol):
                            if isinstance(row[acol], list):
                                self._append_sent[self.tables[tab_index]][acol].update(
                                    set(row[acol])
                                )
                            else:
                                self._append_sent[self.tables[tab_index]][acol].add(
                                    row[acol]
                                )
                        # if there are no new values in the append but there has been a change such that
                        # the pid has ended up in _unsent then send one message to pick up the changes
                        elif row[acol]:
                            try:
                                unsent_appended[acol] = [row[acol][0]]
                            except TypeError:
                                unsent_appended[acol] = [row[acol]]
                    if pid in self._all_sent[tab_index]:
                        message = constructor(
                            self.tables[tab_index],
                            pid,
                            resend=True,
                            unsent_appended=unsent_appended,
                        )
                    else:
                        message = constructor(
                            self.tables[tab_index],
                            pid,
                            unsent_appended=unsent_appended,
                        )
                    if isinstance(message, dict):
                        messages[msg_type].append(message)
                    elif isinstance(message, list):
                        messages[msg_type].extend(message)
                    else:
                        raise TypeError(
                            f"message must be a dictionary or list but was {type(message)}: {message}"
                        )
                self._unsent[tab_index].remove(pid)
                self._sent[tab_index].append(pid)
                self._all_sent[tab_index].append(pid)
        need_to_pop = []
        for key, value in messages.items():
            if not value:
                need_to_pop.append(key)
        for key in need_to_pop:
            messages.pop(key)
        return messages



from __future__ import annotations

import functools
import json
import logging
import math
import os
import pathlib
import queue
import subprocess
import threading
import time
import traceback
from typing import Dict, List, Optional, Set, Tuple

import numpy as np
from gemmi import cif
from pipeliner.api.api_utils import (
    edit_jobstar,
    job_default_parameters_dict,
    write_default_jobstar,
)
from pipeliner.api.manage_project import PipelinerProject
from pipeliner.data_structure import (
    ABORT_FILE,
    FAIL_FILE,
    JOBSTATUS_FAIL,
    JOBSTATUS_RUN,
    SUCCESS_FILE,
)
from pipeliner.job_runner import JobRunner
from pipeliner.pipeliner_job import PipelinerJob
from pipeliner.project_graph import ProjectGraph
from pipeliner.utils import touch

from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions
from relion.pipeline.extra_options import generate_extra_options
from relion.pipeline.options import generate_pipeline_options

logger = logging.getLogger("relion.pipeline")


def wait_for_queued_job_completion(job: PipelinerJob):
    if job.joboptions.get("do_queue") and job.joboptions["do_queue"].get_boolean():
        output_path = pathlib.Path(job.output_dir)
        while not (output_path / SUCCESS_FILE).exists():
            failed = (output_path / FAIL_FILE).exists()
            aborted = (output_path / ABORT_FILE).exists()
            if failed:
                print(f"WARNING: queued job {output_path} failed")
                return
            if aborted:
                print(f"WARNING: queued job {output_path} was aborted")
                return
            time.sleep(10)


def _clear_queue(q: queue.Queue) -> List[str]:
    results = []
    while not q.empty():
        results.append(q.get())
    return results


class PipelineRunner:
    def __init__(
        self,
        projpath: pathlib.Path,
        stopfile: pathlib.Path,
        options: RelionItOptions,
        moviesdir: str = "Movies",
        movietype: str = "mrc",
        restarted: bool = False,
        local: bool = False,
    ):
        self.path = projpath
        self.local = local
        self._restarted = restarted
        self.movies_path = projpath / moviesdir
        self.movietype = movietype if not movietype[0] == "." else movietype[1:]
        self.project = PipelinerProject(make_new_project=True)
        self.stopfile = stopfile
        self.options = options
        self.pipeline_options: Dict[str, dict] = self._generate_pipeline_options()
        self.job_paths: Dict[str, pathlib.Path] = {}
        self.job_objects: Dict[str, PipelinerJob] = {}
        self.job_paths_batch: Dict[str, Dict[str, pathlib.Path]] = {}
        self.job_objects_batch: Dict[str, Dict[str, PipelinerJob]] = {}
        self._past_class_threshold = False
        self._queues: Dict[str, List[queue.Queue]] = {
            "class2D": [queue.Queue()],
            "class3D": [queue.Queue()],
            "ib_group": [queue.Queue()],
        }
        self._passes: List[Set[str]] = [set(), set()]
        self._num_seen_movies = 0
        self._lock = threading.RLock()
        self._extra_options = generate_extra_options
        if self.options.do_second_pass:
            for q in self._queues.values():
                q.append(queue.Queue())
        if restarted:
            self._load_job_paths()
        self._relion_python_exe = os.getenv("RELION_PYTHON_EXECUTABLE")

    def clear_relion_lock(self):
        lock_dir = self.path / ".relion_lock"
        if lock_dir.is_dir():
            for f in lock_dir.glob("*"):
                f.unlink()
            lock_dir.rmdir()

    def _load_job_paths(self):
        search_paths = [("Import", 0), ("MotionCorr", 0)]
        jobs = ["relion.import.movies"]
        if self.options.motioncor_do_own:
            jobs.append("relion.motioncorr.own")
        else:
            jobs.append("relion.motioncorr.motioncor2")
        ib_index = 0
        if self.options.do_icebreaker_job_group:
            search_paths.append(("IceBreaker", ib_index))
            ib_index += 1
            jobs.append("icebreaker.micrograph_analysis.micrographs")
        if self.options.do_icebreaker_job_flatten:
            search_paths.append(("IceBreaker", ib_index))
            ib_index += 1
            jobs.append("icebreaker.micrograph_analysis.enhancecontrast")
        if self.options.do_icebreaker_fivefig:
            search_paths.append(("IceBreaker", ib_index))
            ib_index += 1
            jobs.append("icebreaker.micrograph_analysis.summary")
        search_paths.append(("CtfFind", 0))
        jobs.append("relion.ctffind.ctffind4")

        if not self.options.stop_after_ctf_estimation:
            if self.options.autopick_do_cryolo:
                jobs.append("cryolo.autopick")
            elif self.options.autopick_do_LoG:
                jobs.append("relion.autopick.log")
            search_paths.extend([("AutoPick", 0), ("Extract", 0), ("Select", 0)])
            jobs.extend(["relion.extract", "relion.select.split"])

        for job, sp in zip(jobs, search_paths):
            job_paths = [p for p in (self.path / sp[0]).glob("*") if not p.is_symlink()]
            if not job_paths:
                continue
            ordered_paths = sorted(
                job_paths, key=lambda x: int(x.parts[-1].replace("job", ""))
            )
            self.job_paths[job] = ordered_paths[sp[1]].relative_to(self.path)

        if not self.options.stop_after_ctf_estimation:
            select_job = self.job_paths.get("relion.select.split")
            if not select_job:
                return
            if self.options.do_class2d_vdam:
                class2d_type = "relion.class2d.vdam"
            else:
                class2d_type = "relion.class2d.em"
            self.job_paths_batch[class2d_type] = {}
            self.job_paths_batch["relion.select.class2dauto"] = {}
            self.job_objects_batch[class2d_type] = {}
            self.job_objects_batch["relion.select.class2dauto"] = {}
            for p in (self.path / "Class2D").glob("*"):
                select_file = self._get_select_file(p)
                if select_file:
                    self.job_paths_batch[class2d_type][select_file] = p.relative_to(
                        self.path
                    )
            if self.options.do_icebreaker_group:
                self.job_paths_batch["icebreaker.micrograph_analysis.particles"] = {}
                self.job_objects_batch["icebreaker.micrograph_analysis.particles"] = {}
                for p in (self.path / "IceBreaker").glob("*"):
                    select_file = self._get_select_file(p, option_name="in_parts")
                    if select_file:
                        self.job_paths_batch[
                            "icebreaker.micrograph_analysis.particles"
                        ][select_file] = p.relative_to(self.path)
            if (self.path / "InitialModel").is_dir():
                self.job_paths["relion.initialmodel"] = list(
                    (self.path / "InitialModel").glob("*")
                )[0].relative_to(self.path)
            if (
                self.job_paths.get("cryolo.autopick")
                and (self.path / "AutoPick").is_dir()
                and self.options.estimate_particle_diameter
            ):
                # set the particle diameter from the cryolo job
                self._set_particle_diameter(self.job_paths["cryolo.autopick"])

    def _get_select_file(
        self, class_job_path: pathlib.Path, option_name: str = "fn_img"
    ) -> str:
        try:
            star_doc = cif.read_file(os.fspath(class_job_path / "job.star"))
            data = json.loads(star_doc.as_json())
            joboptions = {
                k: v
                for k, v in zip(
                    data["joboptions_values"]["_rlnjoboptionvariable"],
                    data["joboptions_values"]["_rlnjoboptionvalue"],
                )
            }
            return joboptions[option_name]
        except KeyError:
            return ""

    def _generate_pipeline_options(self):
        if self.local:
            pipeline_jobs = {
                "relion.import.movies": "",
                "relion.motioncorr.motioncor2": "",
                "relion.motioncorr.own": "",
                "icebreaker.micrograph_analysis.micrographs": "",
                "icebreaker.micrograph_analysis.enhancecontrast": "",
                "icebreaker.micrograph_analysis.summary": "",
                "relion.ctffind.ctffind4": "",
                "relion.autopick.log": "",
                "relion.autopick.ref3d": "",
                "cryolo.autopick": "",
                "relion.extract": "",
                "relion.select.split": "",
                "icebreaker.micrograph_analysis.particles": "",
                "relion.class2d.em": "",
                "relion.class2d.vdam": "",
                "relion.initialmodel": "",
                "relion.class3d": "",
            }
        else:
            pipeline_jobs = {
                "relion.import.movies": "",
                "relion.motioncorr.motioncor2": "gpu",
                "relion.motioncorr.own": "cpu",
                "icebreaker.micrograph_analysis.micrographs": "cpu-smp",
                "icebreaker.micrograph_analysis.enhancecontrast": "cpu-smp",
                "icebreaker.micrograph_analysis.summary": "cpu-smp",
                "relion.ctffind.ctffind4": "cpu",
                "relion.autopick.log": "cpu",
                "relion.autopick.ref3d": "cpu",
                "cryolo.autopick": "gpu-smp",
                "relion.extract": "cpu",
                "relion.select.split": "",
                "icebreaker.micrograph_analysis.particles": "cpu-smp",
                "relion.class2d.em": "gpu",
                "relion.class2d.vdam": "gpu-smp",
                "relion.initialmodel": "gpu-smp",
                "relion.class3d": "gpu",
            }
        return generate_pipeline_options(self.options, pipeline_jobs)

    def _do_post_run_actions(self, job: PipelinerJob, project_name: str = "default"):
        # for when a job is submitted to the queue
        # post run actions are only run after the waiting is over
        if job.joboptions.get("do_queue") and job.joboptions["do_queue"].get_boolean():
            output_path = pathlib.Path(job.output_dir)
            try:
                job.post_run_actions()
                post_run_status = JOBSTATUS_RUN
            except Exception as e:
                touch(output_path / FAIL_FILE)
                post_run_status = JOBSTATUS_FAIL

                warn = (
                    f"WARNING: post_run_actions for {output_path} raised an error:\n"
                    f"{str(e)}\n{traceback.format_exc()}"
                )
                with open(output_path / "run.err", "a") as err_file:
                    err_file.write(f"\n{warn}")

            # create default displays for the job's nodes if they are inside the project
            for node in job.output_nodes:
                if node.name[0].isalpha():
                    node.write_default_result_file()

            if self._lock:
                with self._lock:
                    with ProjectGraph(
                        name=project_name, read_only=False
                    ) as post_run_pipeline:
                        job_runner = JobRunner(post_run_pipeline)
                        job_runner.add_job_to_pipeline(job, post_run_status, True)
            else:
                with ProjectGraph(
                    name=project_name, read_only=False
                ) as post_run_pipeline:
                    job_runner = JobRunner(post_run_pipeline)
                    job_runner.add_job_to_pipeline(job, post_run_status, True)

    def fresh_job(
        self,
        job: str,
        extra_params: Optional[dict] = None,
        wait: bool = True,
        lock: Optional[threading.RLock] = None,
        alias: str = "",
    ) -> (PipelinerJob, pathlib.Path):
        logger.info(f"Registering new job: {job}")
        write_default_jobstar(job)
        params = job_default_parameters_dict(job)
        params.update(self.pipeline_options.get(job, {}))
        if extra_params is not None:
            params.update(extra_params)
        _params = {k: str(v) for k, v in params.items() if not isinstance(v, bool)}

        def _b2s(bv: bool) -> str:
            if bv:
                return "Yes"
            return "No"

        _params.update({k: _b2s(v) for k, v in params.items() if isinstance(v, bool)})
        params = _params
        edit_jobstar(
            f"{job.replace('.', '_')}_job.star",
            params,
            f"{job.replace('.', '_')}_job.star",
        )
        if lock is None:
            job_object = self.project.run_job(
                f"{job.replace('.', '_')}_job.star",
                wait_for_queued=wait,
            )
            if alias:
                self.project.set_alias(job_object.output_dir, alias)
        else:
            with lock:
                job_object = self.project.run_job(
                    f"{job.replace('.', '_')}_job.star",
                    wait_for_queued=False,
                )
                if alias:
                    self.project.set_alias(job_object.output_dir, alias)
            wait_for_queued_job_completion(job_object)
            self._do_post_run_actions(job_object)
        logger.info(f"New job registered: {job_object.output_dir}")
        return job_object, pathlib.Path(job_object.output_dir)

    def _get_split_files(self, select_job: pathlib.Path) -> List[str]:
        all_split_files = list(select_job.glob("*particles_split*.star"))
        if len(all_split_files) == 1:
            return [str(all_split_files[0])]
        # drop the most recent batch if there is more than one as it probably isn't complete
        def batch(fname: str) -> int:
            spfname = fname.split("particles_split")
            return int(spfname[-1].replace(".star", ""))

        with_batch_numbers = [(batch(str(f)), str(f)) for f in all_split_files]
        sorted_batch_numbers = sorted(with_batch_numbers, key=lambda x: x[0])
        # if there is more than one batch then we are past the threshold for a single batch
        if len(sorted_batch_numbers) > 1:
            self._past_class_threshold = True
        return [s[1] for s in sorted_batch_numbers[:-1]]

    def _get_num_movies(self, star_file: pathlib.Path) -> int:
        star_doc = cif.read_file(os.fspath(star_file))
        return len(list(star_doc[1].find_loop("_rlnMicrographName")))

    def _set_particle_diameter(self, autopick_job: pathlib.Path):
        # Find the diameter of the biggest particle in cryolo
        cryolo_particle_sizes = np.array([])
        for cbox_file in autopick_job.glob("CBOX/*.cbox"):
            cbox_block = cif.read_file(str(cbox_file)).find_block("cryolo")
            cbox_sizes = np.append(
                np.array(cbox_block.find_loop("_EstWidth"), dtype=float),
                np.array(cbox_block.find_loop("_EstHeight"), dtype=float),
            )
            cbox_confidence = np.append(
                np.array(cbox_block.find_loop("_Confidence"), dtype=float),
                np.array(cbox_block.find_loop("_Confidence"), dtype=float),
            )
            cryolo_particle_sizes = np.append(
                cryolo_particle_sizes,
                cbox_sizes[cbox_confidence > self.options.cryolo_threshold],
            )
        try:
            particle_diameter_pixels = np.quantile(cryolo_particle_sizes, 0.75)
            # Set the new particle diameter in the pipeline options
            self.options.particle_diameter = (
                particle_diameter_pixels * self.options.angpix
            )
            self.pipeline_options = self._generate_pipeline_options()
        except IndexError:
            print(
                "WARNING: Unable to set particle diameter as no particles were picked by cryolo"
            )

    def preprocessing(
        self, ref3d: str = "", ref3d_angpix: float = -1
    ) -> Optional[List[str]]:
        logger.info("Preprocessing started")
        if ref3d and self.options.autopick_do_cryolo:
            return []

        jobs = ["relion.import.movies"]
        aliases = {}
        if self.options.motioncor_do_own:
            jobs.append("relion.motioncorr.own")
        else:
            jobs.append("relion.motioncorr.motioncor2")
        if self.options.do_icebreaker_job_group:
            jobs.append("icebreaker.micrograph_analysis.micrographs")
            aliases["icebreaker.micrograph_analysis.micrographs"] = "Icebreaker_G"
        if self.options.do_icebreaker_job_flatten:
            jobs.append("icebreaker.micrograph_analysis.enhancecontrast")
            aliases["icebreaker.micrograph_analysis.enhancecontrast"] = "Icebreaker_F"
        if self.options.do_icebreaker_fivefig:
            jobs.append("icebreaker.micrograph_analysis.summary")
            aliases["icebreaker.micrograph_analysis.summary"] = "Icebreaker_5fig"
        jobs.append("relion.ctffind.ctffind4")

        for job in jobs:
            if not self.job_paths.get(job):
                self.job_objects[job], self.job_paths[job] = self.fresh_job(
                    job,
                    extra_params=self._extra_options(job, self.job_paths, self.options),
                    lock=self._lock,
                    alias=aliases.get(job, ""),
                )
            else:
                if self._lock:
                    with self._lock:
                        self.job_objects[job] = self.project.continue_job(
                            str(self.job_paths[job]), wait_for_queued=False
                        )
                    wait_for_queued_job_completion(self.job_objects[job])
                    self._do_post_run_actions(self.job_objects[job])
                else:
                    self.project.continue_job(str(self.job_paths[job]))
        if self.job_paths.get("relion.motioncorr.own"):
            self._num_seen_movies = self._get_num_movies(
                self.job_paths["relion.motioncorr.own"] / "corrected_micrographs.star"
            )
        elif self.job_paths.get("relion.motioncorr.motioncor2"):
            self._num_seen_movies = self._get_num_movies(
                self.job_paths["relion.motioncorr.motioncor2"]
                / "corrected_micrographs.star"
            )
        else:
            logger.error(
                "Neither a relion.motioncorr.own nor a relion.motioncorr.motioncor2 job were found"
            )
            raise KeyError(
                "Neither a relion.motioncorr.own nor a relion.motioncorr.motioncor2 job were found"
            )
        if self.options.stop_after_ctf_estimation:
            return []

        if self.options.autopick_do_cryolo:
            next_jobs = ["cryolo.autopick"]
        elif self.options.autopick_do_LoG and not ref3d:
            next_jobs = ["relion.autopick.log"]
        else:
            next_jobs = ["relion.autopick.ref3d"]
        next_jobs.extend(["relion.extract" + ref3d, "relion.select.split" + ref3d])
        for job in next_jobs:
            if not self.job_paths.get(job):
                if job == "relion.autopick.ref3d":
                    try:
                        self.job_objects[job], self.job_paths[job] = self.fresh_job(
                            job.replace(ref3d, ""),
                            extra_params={
                                **self._extra_options(
                                    job, self.job_paths, self.options
                                ),
                                **{
                                    "fn_ref3d_autopick": ref3d,
                                    "angpix_ref": ref3d_angpix,
                                },
                            },
                            lock=self._lock,
                        )
                    except Exception:
                        logger.warning(f"Failed to register fresh job: {job}")
                        return None
                else:
                    try:
                        self.job_objects[job], self.job_paths[job] = self.fresh_job(
                            job.replace(ref3d, ""),
                            extra_params=self._extra_options(
                                job, self.job_paths, self.options
                            ),
                            lock=self._lock,
                        )
                    except Exception:
                        logger.warning(f"Failed to register fresh job: {job}")
                        return None

                if job == "cryolo.autopick" and self.options.estimate_particle_diameter:
                    # set the particle diameter from the output of the first batch
                    self._set_particle_diameter(self.job_paths[job])
            else:
                if self._lock:
                    with self._lock:
                        self.job_objects[job] = self.project.continue_job(
                            str(self.job_paths[job]), wait_for_queued=False
                        )
                    wait_for_queued_job_completion(self.job_objects[job])
                    self._do_post_run_actions(self.job_objects[job])
                else:
                    self.project.continue_job(str(self.job_paths[job]))
        select_path = self.job_paths["relion.select.split" + ref3d]
        return self._get_split_files(select_path)

    @functools.lru_cache(maxsize=1)
    def _best_class(
        self, job: str = "relion.initialmodel", batch: str = ""
    ) -> Tuple[Optional[str], Optional[float]]:
        relion_chosen_ref = ""
        try:
            model_file_candidates = list(
                (self.path / self.job_paths_batch[job][batch]).glob("*_model.star")
            )
        except (TypeError, KeyError):
            if (self.path / self.job_paths[job] / "initial_model.mrc").is_file():
                relion_chosen_ref = str(self.job_paths[job] / "initial_model.mrc")
            model_file_candidates = list(
                (self.path / self.job_paths[job]).glob("*_model.star")
            )

        def iteration_count(x: pathlib.Path) -> int:
            parts = str(x).split("_")
            for _x in parts:
                if _x.startswith("it"):
                    return int(_x.replace("it", ""))
            return 0

        model_file_candidates = sorted(model_file_candidates, key=iteration_count)
        model_file_candidates.reverse()
        model_file = model_file_candidates[0]
        try:
            star_doc = cif.read_file(os.fspath(model_file))
            block_num = None
            for block_index, block in enumerate(star_doc):
                if list(block.find_loop("_rlnReferenceImage")):
                    block_num = block_index
                    break
            if block_num is None:
                return None, None
            star_block = star_doc[block_num]
            ref = list(star_block.find_loop("_rlnReferenceImage"))[0]
            ref_size: float = 0
            ref_resolution = None
            for i, image in enumerate(star_block.find_loop("_rlnReferenceImage")):
                size = float(list(star_block.find_loop("_rlnClassDistribution"))[i])
                resolution = float(
                    list(star_block.find_loop("_rlnEstimatedResolution"))[i]
                )
                if ref_resolution is None or resolution < ref_resolution:
                    ref_resolution = resolution
                    ref_size = size
                    ref = image
                elif resolution == ref_resolution and size > ref_size:
                    ref_resolution = resolution
                    ref_size = size
                    ref = image
            return relion_chosen_ref or ref.split("@")[-1], float(
                star_doc[0].find_value("_rlnPixelSize")
            )
        except Exception as e:
            logger.warning(f"Exception caught: {e}", exc_info=True)
            print(e)
            return None, None

    @functools.lru_cache(maxsize=1)
    def _best_class_fsc(
        self,
        angpix: float,
        boxsize: int,
        job: str = "relion.initialmodel",
        batch: str = "",
    ) -> Tuple[Optional[str], Optional[float]]:
        fsc_files = []
        try:
            model_file_candidates = list(
                (self.path / self.job_paths_batch[job][batch]).glob("*_model.star")
            )
        except (TypeError, KeyError):
            model_file_candidates = list(
                (self.path / self.job_paths[job]).glob("*_model.star")
            )

        def iteration_count(x: pathlib.Path) -> int:
            parts = str(x).split("_")
            for _x in parts:
                if _x.startswith("it"):
                    return int(_x.replace("it", ""))
            return 0

        model_file_candidates = sorted(model_file_candidates, key=iteration_count)
        model_file_candidates.reverse()
        model_file = model_file_candidates[0]
        data_file = pathlib.Path(str(model_file).replace("model", "data"))
        try:
            star_doc = cif.read_file(os.fspath(model_file))
            block_num = None
            for block_index, block in enumerate(star_doc):
                if list(block.find_loop("_rlnReferenceImage")):
                    block_num = block_index
                    break
            if block_num is None:
                return None, None
            star_block = star_doc[block_num]
        except Exception:
            return None, None

        mask_outer_radius = math.floor(0.98 * self.options.mask_diameter / (2 * angpix))
        (
            self.job_objects["relion.external.mask_soft_edge"],
            self.job_paths["relion.external.mask_soft_edge"],
        ) = self.fresh_job(
            "relion.external",
            extra_params={
                "fn_exe": "external_job_mask_soft_edge",
                "param1_label": "box_size",
                "param1_value": str(boxsize),
                "param2_label": "angpix",
                "param2_value": str(angpix),
                "param3_label": "outer_radius",
                "param3_value": str(mask_outer_radius),
            },
            alias="MaskSoftEdge",
            lock=self._lock,
        )
        for iclass in range(1, len(star_block.find_loop("_rlnReferenceImage")) + 1):
            (
                self.job_objects[f"relion.external.select_and_split_{iclass}"],
                self.job_paths[f"relion.external.select_and_split_{iclass}"],
            ) = self.fresh_job(
                "relion.external",
                extra_params={
                    "fn_exe": "external_job_select_and_split",
                    "in_mic": str(data_file.relative_to(self.path)),
                    "param1_label": "in_dir",
                    "param1_value": f"{self.job_paths[job]}",
                    "param2_label": "outfile",
                    "param2_value": f"particles_class{iclass}.star",
                    "param3_label": "class_number",
                    "param3_value": str(iclass),
                },
                alias=f"SelectAndSplit_{iclass}",
                lock=self._lock,
            )
            (
                self.job_objects[f"relion.external.reconstruct_halves_{iclass}"],
                self.job_paths[f"relion.external.reconstruct_halves_{iclass}"],
            ) = self.fresh_job(
                "relion.external",
                extra_params={
                    "fn_exe": "external_job_reconstruct_halves",
                    "in_mic": f"External/SelectAndSplit_{iclass}/particles_class{iclass}.star",
                    "param1_label": "in_dir",
                    "param1_value": f"External/SelectAndSplit_{iclass}",
                    "param2_label": "i",
                    "param2_value": f"particles_class{iclass}.star",
                    "param3_label": "mask_diameter",
                    "param3_value": f"{self.options.mask_diameter}",
                    "param4_label": "class_number",
                    "param4_value": str(iclass),
                    "do_queue": "Yes",
                    "qsubscript": self.options.queue_submission_template_cpu_smp,
                },
                alias=f"ReconstructHalves_{iclass}",
                lock=self._lock,
            )
            (
                self.job_objects[f"relion.postprocess_{iclass}"],
                self.job_paths[f"relion.postprocess_{iclass}"],
            ) = self.fresh_job(
                "relion.postprocess",
                extra_params={
                    "fn_mask": "External/MaskSoftEdge/mask.mrc",
                    "fn_in": f"External/ReconstructHalves_{iclass}/3d_half1_model{iclass}.mrc",
                    "angpix": str(angpix),
                    "do_queue": "Yes",
                    "qsubscript": self.options.queue_submission_template_cpu_smp,
                },
                alias=f"GetFSC_{iclass}",
                lock=self._lock,
            )
            fsc_files.append(f"PostProcess/GetFSC_{iclass}/postprocess.star")

        (
            self.job_objects["relion.external.fsc_fitting"],
            self.job_paths["relion.external.fsc_fitting"],
        ) = self.fresh_job(
            "relion.external",
            extra_params={
                "fn_exe": "external_job_fsc_fitting",
                "param1_label": "i",
                "param1_value": " ".join(fsc_files),
            },
            alias="FSCFitting",
            lock=self._lock,
        )
        with open(
            self.job_paths["relion.external.fsc_fitting"] / "BestClass.txt", "r"
        ) as f:
            class_index = int(f.readline())
        split_ref_img = list(star_block.find_loop("_rlnReferenceImage"))[
            class_index
        ].split("@")
        return (
            split_ref_img[-1],
            float(star_doc[0].find_value("_rlnPixelSize")),
        )

    def _new_movies(self, glob_pattern: str = "") -> bool:
        num_movies = len(
            [
                m
                for m in self.movies_path.glob(glob_pattern or "**/*")
                if m.suffix == "." + self.movietype
            ]
        )
        return not num_movies == self._num_seen_movies

    def _classification_3d(
        self,
        angpix: Optional[float] = None,
        boxsize: Optional[int] = None,
        iteration: int = 0,
        single_job: bool = True,
        clear_queue: bool = True,
    ):
        while True:
            if clear_queue:
                # clear the queue down to the most recent item (doesn't need to function perfectly)
                while self._queues["class3D"][iteration].qsize() > 1:
                    self._queues["class3D"][iteration].get()
                batch_file = self._queues["class3D"][iteration].get()
            else:
                batch_file = self._queues["class3D"][iteration].get()
            if not batch_file:
                return
            if (
                self.job_paths.get("relion.initialmodel") is None
                and not self.options.have_3d_reference
            ):
                (
                    self.job_objects["relion.initialmodel"],
                    self.job_paths["relion.initialmodel"],
                ) = self.fresh_job(
                    "relion.initialmodel",
                    extra_params={"fn_img": batch_file},
                    lock=self._lock,
                )
            if not self.job_paths_batch.get("relion.class3d"):
                self.job_paths_batch["relion.class3d"] = {}
                self.job_objects_batch["relion.class3d"] = {}
            if self.options.use_fsc_criterion and angpix is None:
                logger.error(
                    "use_fsc_criterion is True but angpix has not been specified"
                )
                raise ValueError(
                    "use_fsc_criterion is True but angpix has not been specified"
                )
            try:
                if self.options.have_3d_reference:
                    ref = self.options.class3d_reference
                elif self.options.use_fsc_criterion:
                    ref = self._best_class_fsc(angpix, boxsize)[0]
                else:
                    ref = self._best_class()[0]
                if (
                    not self.job_paths_batch["relion.class3d"].get(batch_file)
                    or not single_job
                ):
                    (
                        self.job_objects_batch["relion.class3d"][batch_file],
                        self.job_paths_batch["relion.class3d"][batch_file],
                    ) = self.fresh_job(
                        "relion.class3d",
                        extra_params={
                            "fn_img": batch_file,
                            "fn_ref": ref,
                        },
                        lock=self._lock,
                    )
                else:
                    if self._lock:
                        try:
                            with self._lock:
                                self.job_objects_batch["relion.class3d"][
                                    batch_file
                                ] = self.project.run_job(
                                    "relion_class3d_job.star",
                                    overwrite=str(
                                        self.job_paths_batch["relion.class3d"][
                                            batch_file
                                        ]
                                    ),
                                    wait_for_queued=False,
                                )
                            wait_for_queued_job_completion(
                                self.job_objects_batch["relion.class3d"][batch_file]
                            )
                            self._do_post_run_actions(
                                self.job_objects_batch["relion.class3d"][batch_file]
                            )
                        except (AttributeError, FileNotFoundError) as e:
                            logger.warning(
                                f"Exception encountered in 3D classification runner. Try again: {e}"
                            )
                            self.clear_relion_lock()
                            continue
                    else:
                        self.project.run_job(
                            "relion_class3d_job.star",
                            overwrite=str(
                                self.job_paths_batch["relion.class3d"][batch_file]
                            ),
                            wait_for_queued=True,
                        )
            except (AttributeError, FileNotFoundError) as e:
                logger.warning(
                    f"Exception encountered in 3D classification runner. Try again: {e}"
                )
                print(
                    f"Exception encountered in 3D classification runner. Try again: {e}"
                )
                self.clear_relion_lock()
                # self._queues["class3D"][iteration].put(batch_file)
            except Exception as e:
                logger.warning(
                    f"Unexpected Exception in 3D classification runner: {e}",
                    exc_info=True,
                )
                return

    def classification(
        self,
        angpix: Optional[float] = None,
        boxsize: Optional[int] = None,
        iteration: int = 0,
    ):
        fraction_of_classes_to_remove = (
            self.options.class2d_fraction_of_classes_to_remove
        )
        files_to_combine = ""
        quantile_threshold = 0
        last_completed_split = 0
        batch_size_3d = 0
        class3d_thread = None
        stop_3d = False
        while True:
            try:
                batch_file, batch_is_complete = self._queues["class2D"][iteration].get()
                if not batch_file:
                    if class3d_thread is None:
                        return
                    self._queues["class3D"][iteration].put("")
                    class3d_thread.join()
                    return
                if batch_file == "__kill__":
                    if class3d_thread is None:
                        return
                    _clear_queue(self._queues["class3D"][iteration])
                    self._queues["class3D"][iteration].put("")
                    class3d_thread.join()
                    return
                if self.options.do_class2d_vdam:
                    class2d_type = "relion.class2d.vdam"
                else:
                    class2d_type = "relion.class2d.em"
                if (
                    self.job_paths_batch.get(class2d_type, {}).get(batch_file)
                    and self._restarted
                ):
                    continue

                if not self.job_paths_batch.get(class2d_type):
                    # if this is the first batch, start a new 2D classification job
                    self.job_paths_batch[class2d_type] = {}
                    self.job_paths_batch["relion.select.class2dauto"] = {}
                    self.job_objects_batch[class2d_type] = {}
                    self.job_objects_batch["relion.select.class2dauto"] = {}
                    try:
                        (
                            self.job_objects_batch[class2d_type][batch_file],
                            self.job_paths_batch[class2d_type][batch_file],
                        ) = self.fresh_job(
                            class2d_type,
                            extra_params={"fn_img": batch_file},
                            lock=self._lock,
                        )
                    except (AttributeError, FileNotFoundError) as e:
                        logger.warning(
                            f"Exception encountered in 2D classification runner. Try again: {e}"
                        )
                        self.clear_relion_lock()
                        continue
                elif self.job_paths_batch[class2d_type].get(batch_file):
                    # runs when a previously incomplete batch is repeated
                    if self._lock:
                        try:
                            with self._lock:
                                self.job_objects_batch[class2d_type][
                                    batch_file
                                ] = self.project.run_job(
                                    f"{class2d_type.replace('.', '_')}_job.star",
                                    overwrite=str(
                                        self.job_paths_batch[class2d_type][batch_file]
                                    ),
                                    wait_for_queued=False,
                                )
                            wait_for_queued_job_completion(
                                self.job_objects_batch[class2d_type][batch_file]
                            )
                            self._do_post_run_actions(
                                self.job_objects_batch[class2d_type][batch_file]
                            )
                        except (AttributeError, FileNotFoundError) as e:
                            logger.warning(
                                f"Exception encountered in 2D classification runner. Try again: {e}"
                            )
                            self.clear_relion_lock()
                            continue
                    else:
                        self.project.run_job(
                            f"{class2d_type.replace('.', '_')}_job.star",
                            overwrite=str(
                                self.job_paths_batch[class2d_type][batch_file]
                            ),
                            wait_for_queued=True,
                        )
                else:
                    # classification for all new batches except the first
                    try:
                        (
                            self.job_objects_batch[class2d_type][batch_file],
                            self.job_paths_batch[class2d_type][batch_file],
                        ) = self.fresh_job(
                            class2d_type,
                            extra_params={"fn_img": batch_file},
                            lock=self._lock,
                        )
                    except (AttributeError, FileNotFoundError) as e:
                        logger.warning(
                            f"Exception encountered in 2D classification runner. Try again: {e}"
                        )
                        self.clear_relion_lock()
                        continue

                split_file = ""
                split_file_column = []
                if batch_is_complete and fraction_of_classes_to_remove:
                    if not self.job_paths_batch.get("relion.select.class2dauto"):
                        # the first time, run 2D class selection to produce rankings
                        (
                            self.job_objects_batch["relion.select.class2dauto"][
                                batch_file
                            ],
                            self.job_paths_batch["relion.select.class2dauto"][
                                batch_file
                            ],
                        ) = self.fresh_job(
                            "relion.select.class2dauto",
                            extra_params={
                                "fn_model": self.job_paths_batch[class2d_type][
                                    batch_file
                                ]
                                / f"run_it{self.options.class2d_nr_iter:03}_optimiser.star",
                                "python_exe": self._relion_python_exe,
                                "other_args": f"--select_min_nr_particles {int(self.options.batch_size / 2)}",
                            },
                            lock=self._lock,
                        )

                    if fraction_of_classes_to_remove == 1:
                        # if fraction to remove is set to 1, select only on particles
                        quantile_threshold = 1
                    if not quantile_threshold:
                        # get class rankings from the first batch
                        quantile_key = list(
                            self.job_objects_batch["relion.select.class2dauto"].keys()
                        )[0]
                        star_doc = cif.read_file(
                            str(
                                self.job_paths_batch["relion.select.class2dauto"][
                                    quantile_key
                                ]
                                / "rank_model.star"
                            )
                        )
                        # find threshold for particle selection
                        star_block = star_doc["model_classes"]
                        class_scores = np.array(
                            star_block.find_loop("_rlnClassScore"), dtype=float
                        )
                        quantile_threshold = np.quantile(
                            class_scores,
                            fraction_of_classes_to_remove,
                        )

                    if self.job_paths_batch["relion.select.class2dauto"].get(
                        batch_file
                    ):
                        # re-run selection for the first batch using the threshold
                        edit_jobstar(
                            str(
                                self.job_paths_batch["relion.select.class2dauto"][
                                    batch_file
                                ]
                                / "job.star"
                            ),
                            {"rank_threshold": quantile_threshold},
                            str(
                                self.job_paths_batch["relion.select.class2dauto"][
                                    batch_file
                                ]
                                / "job.star"
                            ),
                        )
                        if self._lock is None:
                            self.project.continue_job(
                                str(
                                    self.job_paths_batch["relion.select.class2dauto"][
                                        batch_file
                                    ]
                                )
                            )
                        else:
                            with self._lock:
                                self.project.continue_job(
                                    str(
                                        self.job_paths_batch[
                                            "relion.select.class2dauto"
                                        ][batch_file]
                                    )
                                )
                    else:
                        # run 2D class selection for batches after the first
                        (
                            self.job_objects_batch["relion.select.class2dauto"][
                                batch_file
                            ],
                            self.job_paths_batch["relion.select.class2dauto"][
                                batch_file
                            ],
                        ) = self.fresh_job(
                            "relion.select.class2dauto",
                            extra_params={
                                "fn_model": self.job_paths_batch[class2d_type][
                                    batch_file
                                ]
                                / f"run_it{self.options.class2d_nr_iter:03}_optimiser.star",
                                "rank_threshold": quantile_threshold,
                                "python_exe": self._relion_python_exe,
                                "other_args": f"--select_min_nr_particles {int(self.options.batch_size / 2)}",
                            },
                            lock=self._lock,
                        )

                    # add the selected particles to the list of particles to use
                    files_to_combine += str(
                        self.job_paths_batch["relion.select.class2dauto"][batch_file]
                        / "particles.star"
                    )

                    # send particles to the file combiner
                    if not self.job_paths.get("combine_star_files_job"):
                        # if this is the first time then create a new job
                        (
                            self.job_objects["combine_star_files_job"],
                            self.job_paths["combine_star_files_job"],
                        ) = self.fresh_job(
                            "combine_star_files_job",
                            extra_params={
                                "files_to_process": files_to_combine,
                                "do_split": True,
                                "split_size": batch_size_3d or self.options.batch_size,
                            },
                            lock=self._lock,
                        )
                    else:
                        # other batches can be run in the same job, with a new file list
                        changed_params = {
                            "files_to_process": files_to_combine,
                            "split_size": str(batch_size_3d or self.options.batch_size),
                        }
                        edit_jobstar(
                            str(self.job_paths["combine_star_files_job"] / "job.star"),
                            changed_params,
                            str(self.job_paths["combine_star_files_job"] / "job.star"),
                        )
                        if self._lock is None:
                            self.project.continue_job(
                                str(self.job_paths["combine_star_files_job"])
                            )
                        else:
                            with self._lock:
                                self.project.continue_job(
                                    str(self.job_paths["combine_star_files_job"])
                                )
                    # all selected particles will go back into the combiner
                    # needs a space at the end of the string
                    files_to_combine = str(
                        self.job_paths["combine_star_files_job"] / "particles_all.star "
                    )
                    # find the batch to determine its length
                    split_file = str(
                        self.job_paths["combine_star_files_job"]
                        / "particles_split1.star"
                    )
                    split_file_block = cif.read_file(split_file)["particles"]
                    split_file_column = list(
                        split_file_block.find_loop("_rlnCoordinateX")
                    )
                if stop_3d:
                    continue
                if (
                    len(split_file_column)
                    >= (last_completed_split + 1) * self.options.batch_size
                    and files_to_combine
                ) or (batch_is_complete and not fraction_of_classes_to_remove):
                    if (
                        len(split_file_column) == batch_size_3d
                        and fraction_of_classes_to_remove
                    ):
                        stop_3d = True
                    # if the split is complete then run 3D classification
                    if self.options.do_class3d and class3d_thread is None:
                        class3d_thread = threading.Thread(
                            target=self._classification_3d,
                            name="3D_classification_runner",
                            kwargs={
                                "iteration": iteration,
                                "angpix": angpix,
                                "boxsize": boxsize,
                                "single_job": True if files_to_combine else False,
                                "clear_queue": True if files_to_combine else False,
                            },
                        )
                        class3d_thread.start()
                        batch_size_3d = self.options.class3d_max_size
                    self._queues["class3D"][iteration].put(
                        split_file if files_to_combine else batch_file
                    )
                    # mark this split as the last one completed and run
                    last_completed_split += 1
            except Exception as e:
                logger.warning(
                    f"Unexpected Exception in 2D classification runner: {e}",
                    exc_info=True,
                )
                return

    def ib_group(self, iteration: int = 0):
        batch_number = 0
        while True:
            batch_file = self._queues["ib_group"][iteration].get()
            batch_number += 1
            if not batch_file:
                return
            if not self.job_paths_batch.get("icebreaker.micrograph_analysis.particles"):
                self.job_paths_batch["icebreaker.micrograph_analysis.particles"] = {}
                self.job_objects_batch["icebreaker.micrograph_analysis.particles"] = {}
            try:
                (
                    self.job_objects_batch["icebreaker.micrograph_analysis.particles"][
                        batch_file
                    ],
                    self.job_paths_batch["icebreaker.micrograph_analysis.particles"][
                        batch_file
                    ],
                ) = self.fresh_job(
                    "icebreaker.micrograph_analysis.particles",
                    extra_params={
                        "in_mics": str(
                            self.job_paths["icebreaker.micrograph_analysis.micrographs"]
                            / "grouped_micrographs.star"
                        ),
                        "in_parts": batch_file,
                    },
                    lock=self._lock,
                    alias=f"Icebreaker_group_batch_{batch_number}_pass_{iteration+1}"
                    if iteration
                    else f"Icebreaker_group_batch_{batch_number}",
                )
            except (AttributeError, FileNotFoundError) as e:
                logger.warning(
                    f"Exception encountered in IceBreaker runner. Try again: {e}"
                )
                print(f"Exception encountered in IceBreaker runner. Try again: {e}")
                self.clear_relion_lock()
                # self._queues["ib_group"][iteration].put(batch_file)
                # batch_number -= 1
            except Exception as e:
                logger.warning(
                    f"Unexpected Exception in IceBreaker runner: {e}", exc_info=True
                )
                return

    def run(self, timeout: int):
        start_time = time.time()
        current_time = start_time
        class_thread = None
        ib_thread = None
        class_thread_second_pass = None
        ib_thread_second_pass = None
        ref3d: str = ""
        ref3d_angpix: float = 1
        iteration = 0
        old_iteration = 0
        first_batch = ""
        continue_anyway = False
        movie_dir = str(self.movies_path.relative_to(self.path))
        if self.options.import_images.startswith(movie_dir):
            glob_pattern = self.options.import_images[len(movie_dir) :]
        else:
            glob_pattern = self.options.import_images
        if glob_pattern.startswith("/"):
            glob_pattern = glob_pattern[1:]
        while not self.stopfile.exists() and (
            current_time - start_time < timeout
            or continue_anyway
            or self._new_movies(glob_pattern=glob_pattern)
        ):
            if self._new_movies(glob_pattern=glob_pattern) or iteration - old_iteration:
                if iteration - old_iteration:
                    continue_anyway = False
                try:
                    split_files = self.preprocessing(
                        ref3d=ref3d, ref3d_angpix=ref3d_angpix
                    )
                except (AttributeError, FileNotFoundError) as e:
                    logger.warning(
                        f"Exception encountered in preprocessing. Try again: {e}",
                        exc_info=True,
                    )
                    print(f"Exception encountered in preprocessing. Try again: {e}")
                    continue
                if not split_files:
                    continue
                if not first_batch:
                    first_batch = split_files[0]
                if self.options.do_icebreaker_group:
                    if ib_thread is None and not iteration:
                        ib_thread = threading.Thread(
                            target=self.ib_group,
                            name="ib_group_runner",
                            kwargs={"iteration": iteration},
                        )
                        ib_thread.start()
                    elif ib_thread_second_pass is None and iteration:
                        ib_thread_second_pass = threading.Thread(
                            target=self.ib_group,
                            name="ib_group_runner_second_pass",
                            kwargs={"iteration": iteration},
                        )
                        ib_thread_second_pass.start()
                    if self._past_class_threshold:
                        new_batches = [
                            f for f in split_files if f not in self._passes[iteration]
                        ]
                        for sf in new_batches:
                            self._queues["ib_group"][iteration].put(sf)
                        if not self.options.do_class2d:
                            self._passes[iteration].update(new_batches)
                if self.options.do_class2d:
                    if class_thread is None and not iteration:
                        curr_angpix = (
                            self.options.angpix * self.options.motioncor_binning
                        )
                        bcb = int(
                            self.options.extract_boxsize
                            / self.options.motioncor_binning
                        )
                        curr_boxsize = bcb + bcb % 2
                        if self.options.extract_downscale:
                            curr_angpix *= (
                                self.options.extract_boxsize
                                / self.options.extract_small_boxsize
                            )
                            curr_boxsize = self.options.extract_small_boxsize
                        class_thread = threading.Thread(
                            target=self.classification,
                            name="classification_runner",
                            kwargs={
                                "iteration": iteration,
                                "angpix": curr_angpix,
                                "boxsize": curr_boxsize,
                            },
                        )
                        class_thread.start()
                    elif class_thread_second_pass is None and iteration:
                        curr_angpix = (
                            self.options.angpix * self.options.motioncor_binning
                        )
                        bcb = int(
                            self.options.extract_boxsize
                            / self.options.motioncor_binning
                        )
                        curr_boxsize = bcb + bcb % 2
                        if self.options.extract2_downscale:
                            curr_angpix *= (
                                self.options.extract_boxsize
                                / self.options.extract2_small_boxsize
                            )
                            curr_boxsize = self.options.extract2_small_boxsize
                        class_thread_second_pass = threading.Thread(
                            target=self.classification,
                            name="classification_runner_second_pass",
                            kwargs={
                                "iteration": iteration,
                                "angpix": curr_angpix,
                                "boxsize": curr_boxsize,
                            },
                        )
                        class_thread_second_pass.start()
                    if len(split_files) == 1:
                        if split_files[0] not in self._passes[iteration]:
                            self._queues["class2D"][iteration].put(
                                (split_files[0], self._past_class_threshold)
                            )
                            if self._past_class_threshold:
                                self._passes[iteration].update(split_files)
                    else:
                        new_batches = [
                            f for f in split_files if f not in self._passes[iteration]
                        ]
                        if not self._past_class_threshold:
                            self._past_class_threshold = True
                        for sf in new_batches:
                            self._queues["class2D"][iteration].put(
                                (sf, self._past_class_threshold)
                            )
                        self._passes[iteration].update(new_batches)
                old_iteration = iteration
                if (
                    self.options.do_second_pass
                    and self.options.do_class2d
                    and self.options.do_class3d
                    and not self.options.autopick_do_cryolo
                    and not iteration
                ):
                    self._queues["ib_group"][0].put("")
                    self._queues["class2D"][0].put(("", ""))
                    if ib_thread:
                        ib_thread.join()
                    if class_thread:
                        class_thread.join()
                    _ref3d, _ref3d_angpix = self._best_class(
                        "relion.class3d", first_batch
                    )
                    ref3d = _ref3d or ""
                    ref3d_angpix = _ref3d_angpix or ref3d_angpix
                    new_angpix = self.options.angpix * self.options.motioncor_binning
                    if self.options.extract2_downscale:
                        new_angpix *= (
                            self.options.extract_boxsize
                            / self.options.extract2_small_boxsize
                        )

                    if abs(new_angpix - float(self.options.autopick_ref_angpix)) > 1e-3:
                        command = [
                            "relion_image_handler",
                            "--i",
                            str(ref3d),
                            "--o",
                            str("ref3d"),
                            "--angpix",
                            str(ref3d_angpix),
                            "--rescale_angpix",
                            str(new_angpix),
                            "--new_box",
                            str(
                                self.options.extract2_small_boxsize
                                if self.options.extract2_downscale
                                else self.options.extract_boxsize
                            ),
                        ]
                        ref3d = ref3d.replace(".mrcs", "_ref3d.mrcs")
                        subprocess.run(command)
                    iteration = 1
                    continue_anyway = True

            time.sleep(10)
            current_time = time.time()
        logger.info(
            "Stopping processing: current time: {current_time}, start time: {start_time}, timeout: {timeout}"
        )
        if ib_thread is not None:
            logger.info("Stopping IceBreaker thread")
            self._queues["ib_group"][0].put("")
        if class_thread is not None:
            logger.info("Stopping classification thread")
            self._queues["class2D"][0].put(("", ""))
        if ib_thread is not None:
            ib_thread.join()
            logger.info("IceBreaker thread stopped")
        if class_thread is not None:
            class_thread.join()
            logger.info("Classification thread stopped")



from __future__ import annotations

from pathlib import Path

from pipeliner.data_structure import SELECT_DIR
from pipeliner.job_options import BooleanJobOption, IntJobOption, StringJobOption
from pipeliner.nodes import NODE_PARTICLESDATA, Node
from pipeliner.pipeliner_job import ExternalProgram, PipelinerJob

COMBINE_STAR_NAME = "combine_star_files.py"


class ProcessStarFiles(PipelinerJob):
    PROCESS_NAME = "combine_star_files_job"
    OUT_DIR = SELECT_DIR

    def __init__(self):
        super().__init__()

        self.jobinfo.display_name = "Particle star file merging"
        self.jobinfo.short_desc = "Combine and split star files of particles"
        self.jobinfo.long_desc = (
            "Combine star files of particles, then optionally split them again."
        )

        self.jobinfo.programs = [ExternalProgram(command=COMBINE_STAR_NAME)]

        self.joboptions["files_to_process"] = StringJobOption(
            label="The star files from which to combine particles",
            is_required=True,
            help_text="The names of the star files, separated by spaces.",
        )
        self.joboptions["do_split"] = BooleanJobOption(
            label="Whether to split the combined star file", default_value=False
        )
        self.joboptions["n_files"] = IntJobOption(
            label="Number of files to split the combined file into",
            default_value=-1,
            help_text=(
                "Provide either the number of files to split into,"
                " or the number of particles per file."
            ),
            deactivate_if=[("do_split", "is", "False")],
        )
        self.joboptions["split_size"] = IntJobOption(
            label="Number of particles to put in each split",
            default_value=-1,
            help_text=(
                "Provide either the number of files to split into,"
                " or the number of particles per file."
            ),
            deactivate_if=[("do_split", "is", "False")],
        )

        self.set_joboption_order(
            ["files_to_process", "do_split", "n_files", "split_size"]
        )

    def get_commands(self):
        """Construct the command for combining and splitting star files"""
        command = [COMBINE_STAR_NAME]
        file_list = self.joboptions["files_to_process"].get_string().split(" ")
        command.extend(file_list)
        command.extend(["--output_dir", str(self.output_dir)])

        if self.joboptions["do_split"].get_boolean():
            command.extend(["--split"])

            if (
                self.joboptions["n_files"].get_number() <= 0
                and self.joboptions["split_size"].get_number() <= 0
            ):
                raise ValueError(
                    "ERROR: When splitting the combined STAR file into subsets,"
                    " set n_files or split_size to a positive value"
                )

            if self.joboptions["n_files"].get_number() > 0:
                command.extend(["--n_files", self.joboptions["n_files"].get_string()])

            if self.joboptions["split_size"].get_number() > 0:
                command.extend(
                    ["--split_size", self.joboptions["split_size"].get_string()]
                )

        # Add files as input nodes, as long as they are not also the output node
        for particle_file in file_list:
            if particle_file != self.output_dir + "particles_all.star":
                self.input_nodes.append(Node(particle_file, NODE_PARTICLESDATA))
        self.output_nodes.append(
            Node(self.output_dir + "particles_all.star", NODE_PARTICLESDATA)
        )

        return [command]

    def post_run_actions(self):
        """Find any output files produced by the splitting"""
        output_files = Path(self.output_dir).glob("particles_split*.star")
        for split in output_files:
            self.output_nodes.append(Node(str(split), NODE_PARTICLESDATA))



from __future__ import annotations

import argparse
from math import ceil
from pathlib import Path
from typing import List

import pandas as pd
import starfile


def combine_star_files(files_to_process: List[Path], output_dir: Path):
    """Combines any number of particle star files into a single file.

    Parameters:
    files_to_process: A list of the particle star files to combine
    output_dir: The directory in which to save the combined "particles_all.star" file
    """
    final_star_file = {}
    number_of_star_files = 0
    for split_file in files_to_process:
        number_of_star_files += 1
        star_dictionary = starfile.read(split_file)

        data_optics = star_dictionary["optics"]
        data_particles = star_dictionary["particles"]

        try:
            # check that the files have the same optics tables
            if final_star_file["optics"].ne(data_optics).any().any():
                raise IndexError(
                    "Cannot combine star files with different optics tables."
                )
        except KeyError:
            # if this is the first file, construct a new table
            final_star_file["optics"] = data_optics

        try:
            # combine the particle tables
            final_star_file["particles"] = pd.concat(
                (final_star_file["particles"], data_particles)
            )
        except KeyError:
            # if this is the first file, construct a new table
            final_star_file["particles"] = data_particles

    starfile.write(final_star_file, output_dir / "particles_all.star", overwrite=True)

    print(f"Combined {number_of_star_files} files into particles_all.star")


def split_star_file(
    file_to_process: Path,
    output_dir: Path,
    number_of_splits: int or None = None,
    split_size: int or None = None,
):
    """Splits a star file into subfiles.

    The number of subfiles can be given with number_of_splits
    or is determined by split_size, the number of particles for in each file.
    """
    star_dictionary = starfile.read(file_to_process)

    data_optics = star_dictionary["optics"]
    data_particles = star_dictionary["particles"]
    number_of_particles = len(data_particles)

    # determine the number of files and size of the splits
    if number_of_splits:
        if split_size:
            print(
                "Warning: "
                "Both number_of_splits and split_size have been given, "
                "using number_of_splits.",
            )
        split_size = ceil(number_of_particles / number_of_splits)
    elif split_size:
        number_of_splits = ceil(number_of_particles / split_size)
    else:
        raise KeyError("Either number_of_splits or split_size must be given.")

    # create dictionaries of particle data and write these to the new files
    for split in range(number_of_splits):
        particles_to_use = data_particles[split_size * split : split_size * (split + 1)]
        dictionary_to_write = {"optics": data_optics, "particles": particles_to_use}
        starfile.write(
            dictionary_to_write,
            output_dir / f".particles_split{split+1}_tmp.star",
            overwrite=True,
        )
        (output_dir / f".particles_split{split+1}_tmp.star").rename(
            output_dir / f"particles_split{split+1}.star"
        )

    print(
        f"Split {number_of_particles} particles into "
        f"{number_of_splits} files with {split_size} particles in each."
    )


def create_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "files_to_process",
        type=Path,
        nargs="+",
        help="The star files from which to combine particles.",
    )
    parser.add_argument(
        "--output_dir",
        dest="output_dir",
        type=Path,
        help="Folder in which to save the new star files.",
    )

    parser.add_argument(
        "--split",
        dest="do_split",
        action="store_true",
        default=False,
        help="Whether to split the particles again into new star files.",
    )
    parser.add_argument(
        "--n_files",
        dest="n_files",
        type=int,
        default=None,
        help="Number of files to split the particles into.",
    )
    parser.add_argument(
        "--split_size",
        dest="split_size",
        type=int,
        default=None,
        help="Number of particles to write per file.",
    )

    return parser


def main():
    arg_parser = create_parser()
    run_args = vars(arg_parser.parse_args())

    combine_star_files(run_args["files_to_process"], run_args["output_dir"])

    if run_args["do_split"]:
        split_star_file(
            run_args["output_dir"] / "particles_all.star",
            run_args["output_dir"],
            run_args["n_files"],
            run_args["split_size"],
        )



from __future__ import annotations

import json
from pathlib import Path

import procrunner
import workflows.recipe
from pydantic import BaseModel, Field
from pydantic.error_wrappers import ValidationError
from workflows.services.common_service import CommonService


class CryoloParameters(BaseModel):
    boxsize: int
    pix_size: float
    input_path: str = Field(..., min_length=1)
    output_path: str = Field(..., min_length=1)
    config_file: str = "/dls_sw/apps/EM/crYOLO/phosaurus_models/config.json"
    weights: str = (
        "/dls_sw/apps/EM/crYOLO/phosaurus_models/gmodel_phosnet_202005_N63_c17.h5"
    )
    threshold: float = 0.3
    mc_uuid: int
    cryolo_command: str


class CrYOLO(CommonService):
    """
    A service that runs crYOLO particle picking
    """

    # Human readable service name
    _service_name = "CrYOLO"

    # Logger name
    _logger_name = "relion.zocalo.cryolo"

    # Values to extract for ISPyB
    number_of_particles: int

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("crYOLO service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "cryolo",
            self.cryolo,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_cryolo_output(self, line: str):
        """
        Read the output logs of cryolo to determine
        the number of particles that are picked
        """
        if not line:
            return

        if "particles in total has been found" in line:
            line_split = line.split()
            self.number_of_particles += int(line_split[0])

        if line.startswith("Deleted"):
            line_split = line.split()
            self.number_of_particles -= int(line_split[1])

    def cryolo(self, rw, header: dict, message: dict):
        """
        Main function which interprets received messages, runs cryolo
        and sends messages to the ispyb and image services
        """

        class RW_mock:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = RW_mock()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        # Reset number of particles
        self.number_of_particles = 0

        try:
            if isinstance(message, dict):
                cryolo_params = CryoloParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                cryolo_params = CryoloParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError):
            self.log.warning(
                f"crYOLO parameter validation failed for message: {message} "
                + "and recipe parameters: "
                + f"{rw.recipe_step.get('parameters', {})}"
            )
            rw.transport.nack(header)
            return

        # Make a cryolo config file with the correct box size
        with open(cryolo_params.config_file, "r") as json_file:
            data = json.load(json_file)
            data["model"]["anchors"] = [cryolo_params.boxsize, cryolo_params.boxsize]
        with open("config.json", "w") as outfile:
            json.dump(data, outfile)

        # Construct a command to run cryolo with the given parameters
        command = cryolo_params.cryolo_command.split()
        command.extend((["--conf", "config.json"]))

        cryolo_flags = {
            "weights": "--weights",
            "input_path": "-i",
            "output_path": "-o",
            "threshold": "--threshold",
            "cryolo_gpus": "--gpu",
        }

        for k, v in cryolo_params.dict().items():
            if v and (k in cryolo_flags):
                if type(v) is tuple:
                    command.extend((cryolo_flags[k], " ".join(str(_) for _ in v)))
                else:
                    command.extend((cryolo_flags[k], str(v)))

        self.log.info(
            f"Input: {cryolo_params.input_path} "
            + f"Output: {cryolo_params.output_path}"
        )

        # Run cryolo and confirm it ran successfully
        result = procrunner.run(
            command=command, callback_stdout=self.parse_cryolo_output
        )
        if result.returncode:
            self.log.error(
                f"crYOLO failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Extract results for ispyb
        ispyb_parameters = {
            "particle_picking_template": cryolo_params.weights,
            "particle_diameter": cryolo_params.pix_size * cryolo_params.boxsize / 10,
            "number_of_particles": self.number_of_particles,
            "summary_image_full_path": cryolo_params.output_path
            + "/picked_particles.mrc",
        }

        # Forward results to ISPyB
        ispyb_parameters.update(
            {
                "ispyb_command": "buffer",
                "buffer_lookup": {"motion_correction_id": cryolo_params.mc_uuid},
                "buffer_command": {"ispyb_command": "insert_particle_picker"},
            }
        )
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb", ispyb_parameters)

        # Extract results for images service
        with open(
            Path(cryolo_params.output_path + "/STAR/")
            / Path(cryolo_params.input_path).with_suffix(".star").name,
            "r",
        ) as coords_file:
            coords = [line.split() for line in coords_file][6:]
            coords_file.close()

        # Forward results to images service
        self.log.info("Sending to images service")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                destination="images",
                message={
                    "parameters": {"images_command": "picked_particles"},
                    "file": cryolo_params.input_path,
                    "coordinates": coords,
                    "angpix": cryolo_params.pix_size,
                    "diameter": cryolo_params.pix_size * cryolo_params.boxsize,
                    "outfile": cryolo_params.output_path + "/picked_particles.jpeg",
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "parameters": {"images_command": "picked_particles"},
                    "file": cryolo_params.input_path,
                    "coordinates": coords,
                    "angpix": cryolo_params.pix_size,
                    "diameter": cryolo_params.pix_size * cryolo_params.boxsize,
                    "outfile": cryolo_params.output_path + "/picked_particles.jpeg",
                },
            )

        rw.transport.ack(header)



from __future__ import annotations

import argparse
import os

import yaml

from relion.cryolo_relion_it import dls_options
from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--file",
        help="Print options to file. Default is relion_it_options.py. Extension should be .py",
        dest="options_file",
        default="relion_it_options.py",
    )
    parser.add_argument(
        "--gpus",
        help="Number of GPUs available on a local system",
        dest="gpus",
        type=int,
        default=0,
    )
    args = parser.parse_args()
    opts = RelionItOptions()
    opts.update_from(vars(dls_options))
    if os.getenv("RELION_CLUSTER_CONFIG"):
        with open(os.getenv("RELION_CLUSTER_CONFIG"), "r") as config:
            cluster_config = yaml.safe_load(config)
        opts.update_from(cluster_config)
    if args.gpus:
        opts.motioncor_gpu = ":".join(str(i) for i in range(args.gpus))
        opts.motioncor_mpi = args.gpus
        opts.refine_gpu = ":".join(str(i) for i in range(args.gpus))
        opts.refine_mpi = args.gpus + 1
        opts.extract_mpi = 4
        opts.ctffind_mpi = 4
    opts.print_options(out_file=open(args.options_file, "w"))



from __future__ import annotations

import functools
import uuid

from relion.node.environment import Environment


@functools.total_ordering
class Node:
    """
    A basic class to hold data and a function for the processing of
    that data. To allow data to be passed between nodes lists of parent
    and child nodes are kept in _in and _out.
    """

    def __init__(self, name, independent=False, **kwargs):
        self._name = name
        self.nodeid = str(uuid.uuid4())[:8]
        self._out = []
        self._in = []
        self._completed = []
        self.environment = Environment()
        self._link_traffic = {}
        self._share_traffic = {}
        self._append_traffic = {}
        self._call_count = 0
        self._in_multi_call = False
        self._can_append = independent
        self.shape = "oval"
        for key, value in kwargs.items():
            self.environment[key] = value

    def __eq__(self, other):
        if isinstance(other, Node):
            if self.name == other.name and len(self._out) == len(other._out):
                for n in self._out:
                    if n not in other._out:
                        return False
                return True
        return False

    def __hash__(self):
        return hash(("relion.node.Node", self._name, tuple(self._out), self.nodeid))

    def __repr__(self):
        return f"Node({repr(str(self._name))})"

    def __iter__(self):
        return iter(self._out)

    def __len__(self):
        return len(self._out)

    def __lt__(self, other):
        if self._is_child(other):
            return True
        return False

    def __rshift__(self, other):
        if isinstance(other, Node):
            self.link_to(other)

    def __call__(self, *args, **kwargs):
        res = []
        self.environment.load_iterator()
        incomplete = self.environment.step()
        self._in_multi_call = True
        while incomplete:
            curr_res = self.func(*args, **kwargs)
            if isinstance(curr_res, list):
                res.extend(curr_res)
            else:
                res.append(curr_res)
            incomplete = self.environment.step()
        self._in_multi_call = False
        for node in self._out:
            node._completed.append(self)
        self._call_count += 1
        if len(res) == 0:
            return
        if len(res) == 1:
            return res[0]
        return res

    def __getitem__(self, key):
        return self.environment[key]

    def func(self, *args, **kwargs):
        """
        Function called when the Node is called.
        Does nothing for a Node: should be overwritten in derived classes
        """
        pass

    @property
    def name(self):
        return str(self._name)

    def change_name(self, new_name):
        self._name = new_name

    def link_to(
        self,
        next_node,
        traffic=None,
        result_as_traffic: bool = False,
        share: list = None,
    ):
        """
        Link to another Node. The Node being linked to needs to be a child of this Node.
        A data transfer between the nodes can be specified. This transfer will only happen
        if the nodes are placed within a Graph and the Graph is called.

        Keyword arguments:
        next_node -- the Node being linked to
        traffic -- a dict or list of dicts to be passed as a message to next_node (default None)
        result_as_traffic -- use the return value of calling this Node as traffic to next_node (default False)
        share -- list of tuples (key01, key02) of keys specifying the passing of a value in this node's environment with
        key01 to the environment of next_node with key02. This data transfer happens after this node is called
        and the data transferred may therefore be affected by the call (default None)
        """
        if next_node not in self._out:
            self._out.append(next_node)
            next_node._in.append(self)
            if traffic is None:
                self._link_traffic[next_node.nodeid] = {}
            else:
                self._link_traffic[next_node.nodeid] = traffic
            if result_as_traffic:
                self._link_traffic[next_node.nodeid] = None
            if share is not None:
                self._share_traffic[next_node.nodeid] = share

    def propagate(self, share):
        """
        Propagate data to downstream nodes.

        Keyword arguments:
        share -- tuple (key01, key02), the value in this node's environment with key01 will be placed
        into the environment of all child nodes (and their children) with key02
        """
        self.environment.propagate.update({share[1]: self.environment[share[0]]})

    def unlink_from(self, next_node):
        """
        Remove the link to a node

        Keyword arguments:
        next_node -- node in this node's _out that is being unlinked
        """
        if next_node in self._out:
            self._out.remove(next_node)

    def _is_child_checker(self, possible_child, checks):
        if self == possible_child:
            checks.extend([True])
        for child in self:
            checks.extend(child._is_child_checker(possible_child, checks=checks))
        return checks

    def _is_child(self, possible_child):
        if True in self._is_child_checker(possible_child, checks=[]):
            return True
        else:
            return False



from __future__ import annotations

import argparse
import pathlib
import platform
import subprocess
import webbrowser

from relion import Project


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--project", required=True, dest="project")
    args = parser.parse_args()
    relion_dir = pathlib.Path(args.project)
    proj = Project(relion_dir)
    proj.show_job_nodes()
    if platform.system() == "Linux" and "WSL" in platform.release():
        subprocess.run(
            ["wslview", relion_dir / "Pipeline" / "relion_pipeline_jobs.gv.svg"],
            check=True,
        )
    else:
        webbrowser.open(relion_dir / "Pipeline" / "relion_pipeline_jobs.gv.svg")


if __name__ == "__main__":
    run()



from __future__ import annotations

import functools


def update_append(d01, d02):
    for key, value in d02.items():
        if key in d01:
            if not isinstance(d01[key], list):
                d01[key] = [d01[key]]
            d01[key].append(value)
        else:
            d01[key] = value


class Propagate:
    def __init__(self):
        self.store = {}
        self.released = False

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        self.store[key] = value
        if not self.released:
            self.released = True

    def keys(self):
        if self.released:
            return self.store.keys()
        else:
            return []

    def update(self, in_dict):
        self.store.update(in_dict)
        if not self.released:
            self.released = True


class Escalate:
    def __init__(self):
        self.store = {}
        self.released = False

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        if self.released:
            self.store[key] = value

    def start(self, esc):
        self.store = esc
        self.released = True


class Iterate:
    def __init__(self, initial_store):
        self.appended = []
        self.store = initial_store

    def squash(self):
        if self.appended:
            squashed_appended = []
            for a in self.appended:
                squashed_appended.extend(a)
            if (
                len(squashed_appended) != len(self.store)
                and len(self.store)
                and self.store != ["__do not iterate__"]
            ):
                raise ValueError(
                    f"Attempted to update ProtoNode Environment with concatenated lists (from updating with can_append_list option) that was a different size to the pre-existing iterator: {len(squashed_appended)} vs. {len(self.store)}, {list(self.store)}"
                )
            if self.store != ["__do not iterate__"]:
                for i, tr in enumerate(squashed_appended):
                    self.store[i].update(tr)
            else:
                self.store = squashed_appended
            self.appended = []

    def __iter__(self):
        return iter(self.store)

    def update(self, u, can_append_list=False):
        if isinstance(u, dict):
            for s in self.store:
                s.update(u)
        if isinstance(u, list):
            if can_append_list:
                self.appended.append(u)
                return
            if (
                self.store == [{}]
                or self.store == []
                or self.store == ["__do not iterate__"]
            ):
                self.store = u
                return
            if len(u) != len(self.store):
                raise ValueError(
                    f"Attempted to update ProtoNode Environment with a list that was a different size to the pre-existing iterator: {len(u)} vs. {len(self.store)}"
                )
            for i, tr in enumerate(u):
                self.store[i].update(tr)


class Environment:
    def __init__(self, base=None):
        set_base(base, self)
        self.propagate = Propagate()
        self.escalate = Escalate()
        self.temp = {}

    def __getitem__(self, key):
        if key in self.base.keys():
            return self.base[key]
        if key in self.temp.keys():
            return self.temp[key]
        if self.propagate.released:
            if key in self.propagate.keys():
                return self.propagate[key]
        if self.escalate.released:
            return self.escalate[key]

    def __setitem__(self, key, value):
        if key in self.base.keys():
            self.base[key] = value
            return
        if key in self.temp.keys():
            self.temp[key] = value
            return
        self.base[key] = value

    def step(self):
        try:
            self.temp = next(self.iterator)
            if self.temp == {}:
                self.empty = True
            else:
                self.empty = False
            if self.temp == "__do not iterate__":
                self.temp = {}
            return True
        except StopIteration:
            self.reset()
            return False

    def get(self, key):
        return self[key]

    def update(self, traffic, append=False, can_append_list=False):
        if isinstance(traffic, dict):
            if append:
                update_append(self.base, traffic)
            else:
                self.base.update(traffic)
            return
        elif isinstance(traffic, list):
            self.iterate.update(traffic, can_append_list=can_append_list)

    def load_iterator(self):
        self.iterate.squash()
        self.iterator = iter(self.iterate)

    def set_escalate(self, esc):
        self.escalate.start(esc)

    def update_prop(self, prop):
        self.propagate.update(prop)

    def reset(self):
        self.iterate = Iterate(["__do not iterate__"])


@functools.singledispatch
def set_base(base, env: Environment):
    raise TypeError(
        "For a ProtoNode Environment the base must be a dictionary or a list"
    )


@set_base.register(type(None))
def _(base: type(None), env: Environment):
    env.base = {}
    env.iterate = Iterate(["__do not iterate__"])


@set_base.register(dict)
def _(base: dict, env: Environment):
    env.base = base
    env.iterate = Iterate(["__do not iterate__"])


@set_base.register(list)
def _(base: list, env: Environment):
    env.base = {}
    env.iterate = Iterate(base)



#!/usr/bin/env python3
"""
External job to select the micrographs for a chosen class and then split
into two random halves
"""

from __future__ import annotations

import argparse
import os
import os.path
import subprocess

RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"


def run_job(project_dir, job_dir, in_dir, starin, starout, args_list):
    parser = argparse.ArgumentParser()
    parser.add_argument("--class_number", dest="class_number")
    parser.add_argument(
        "--j", dest="threads", help="Number of threads to run (ignored)"
    )
    args = parser.parse_args(args_list)
    if not starout.endswith(".star"):
        raise ValueError(
            "Output file for external SelectAndSplit job must end with .star"
        )

    command = [
        "relion_star_handler",
        "--i",
        f"{starin}",
        "--o",
        f"{os.path.join(job_dir, starout)}",
        "--select",
        "rlnClassNumber",
        f"{args.class_number}",
    ]

    subprocess.run(command, check=True)

    command = [
        "relion_star_handler",
        "--i",
        f"{os.path.join(job_dir, starout)}",
        "--o",
        f"{os.path.join(job_dir, starout)}",
        "--split",
        "--random_order",
        "--nr_split",
        "2",
    ]

    subprocess.run(command, check=True)


def main():
    """Change to the job working directory, then call run_job()"""
    parser = argparse.ArgumentParser()
    parser.add_argument("--in_dir", dest="in_dir", help="Input directory name")
    parser.add_argument(
        "--o", "--out_dir", dest="out_dir", help="Output directory name"
    )
    parser.add_argument("--in_mics", dest="starin", help="Input star file name")
    parser.add_argument("--outfile", dest="starout", help="Output star file name")
    # parser.add_argument("--of", help="Output star file name")
    parser.add_argument(
        "--pipeline_control", help="Directory for pipeline control files"
    )
    known_args, other_args = parser.parse_known_args()
    project_dir = os.getcwd()
    os.makedirs(known_args.out_dir, exist_ok=True)
    os.chdir(known_args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        print(
            " select_and_split_external_job: Removing previous failure indicator file"
        )
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        print(
            " select_and_split_external_job: Removing previous success indicator file"
        )
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        os.chdir("../..")
        run_job(
            project_dir,
            known_args.out_dir,
            known_args.in_dir,
            known_args.starin,
            known_args.starout,
            other_args,
        )
    except Exception:
        os.chdir(known_args.out_dir)
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        os.chdir(known_args.out_dir)
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



from __future__ import annotations

import functools
import itertools
import re

from ispyb import sqlalchemy

# if we replace uuid with count do not include 0 in the count because it will break some bool checks for None

WrapperID = itertools.count(1)


class Table:
    def __init__(
        self,
        columns,
        primary_key,
        unique=None,
        counters=None,
        append=None,
        required=None,
    ):
        self.columns = columns
        self._tab = {c: [] for c in self.columns}
        self._primary_key = primary_key
        self._last_update = {self: 0}
        if unique is None:
            self._unique = None
        else:
            if isinstance(unique, str):
                self._unique = [unique]
            else:
                self._unique = unique
        self._unique = self._make_list(unique, default=None)
        self._counters = self._make_list(counters, default=[])
        self._append = self._make_list(append, default=[])
        self._required = self._make_list(required, default=[])

    def __getitem__(self, key):
        return self._tab[key]

    @staticmethod
    def _make_list(elem, *, default):
        if elem is None:
            return default
        elif isinstance(elem, list):
            return elem
        return [elem]

    def add_row(self, row):
        for req in self._required:
            if row.get(req) is None:
                return None

        modified = False

        unique_check = self._unique_check(row)

        prim_key_arg = unique_check or row.get(self._primary_key)
        if unique_check is None or prim_key_arg not in self._tab[self._primary_key]:
            try:
                for counter in self._counters:
                    row[counter] = len(self._tab[counter]) + 1
            except TypeError:
                pass
        else:
            for counter in self._counters:
                index = self._tab[self._primary_key].index(prim_key_arg)
                row[counter] = self._tab[counter][index]

        # if no primary key is specified and the uniqueness check has not returned one then add the row
        # with a new primary key id
        if prim_key_arg is None or prim_key_arg not in self._tab[self._primary_key]:
            modified = True
            for c in self.columns:
                if c == self._primary_key:
                    self._tab[c].append(prim_key_arg or next(WrapperID))
                else:
                    self._tab[c].append(row.get(c))
        # otherwise use existing primary key
        else:
            index = self._tab[self._primary_key].index(prim_key_arg)
            for c in self.columns:
                if c != self._primary_key:
                    row_value = row.get(c)
                    if self._tab[c][index] != row_value:
                        # if column c is marked for appending to then append to it:
                        # if current value is a list and the new row value is a list, set, or a tuple
                        # then combine them as sets and cast back to list for json serialisation,
                        # otherwise add the new value as an item to the set and cast to list
                        if c in self._append:
                            if row_value is not None:
                                if not isinstance(self._tab[c][index], list):
                                    if self._tab[c][index] is None:
                                        self._tab[c][index] = []
                                    else:
                                        self._tab[c][index] = [self._tab[c][index]]
                                curr_as_set = set(self._tab[c][index])
                                curr_orig = set(self._tab[c][index])
                                if isinstance(row_value, (list, set, tuple)):
                                    curr_as_set |= set(row_value)
                                    self._tab[c][index] = list(curr_as_set)
                                else:
                                    curr_as_set.add(row_value)
                                    self._tab[c][index] = list(curr_as_set)
                                if curr_orig.symmetric_difference(
                                    set(self._tab[c][index])
                                ):
                                    modified = True
                        else:
                            modified = True
                            self._tab[c][index] = row_value

        if modified:
            if prim_key_arg is None:
                return row.get(self._primary_key) or self._tab[self._primary_key][-1]
            else:
                return prim_key_arg
        else:
            return None

    # check if the row being added has already existing values for the columns marked as unique
    def _unique_check(self, in_values):
        try:
            unique_indices = []
            for u in self._unique:
                if in_values.get(u) in self._tab[u]:
                    indices = [
                        i
                        for i, element in enumerate(self._tab[u])
                        if element == in_values.get(u)
                    ]
                    if isinstance(indices, list):
                        unique_indices.append(indices)
                    else:
                        unique_indices.append([indices])
                else:
                    break
            else:
                for i1, ui1 in enumerate(unique_indices):
                    for i2, ui2 in enumerate(unique_indices):
                        if i1 != i2:
                            if set(ui1).isdisjoint(ui2):
                                return None
                overlap_list = unique_indices[0]
                for i in range(len(unique_indices) - 1):
                    curr_overlap = set(unique_indices[i]).intersection(
                        unique_indices[i + 1]
                    )
                    overlap_list = list(set(overlap_list).intersection(curr_overlap))
                if not overlap_list:
                    return None
                return self._tab[self._primary_key][overlap_list[0]]
            return None
        except TypeError:
            return None

    def get_row_index(self, key, value):
        if value is None:
            return None
        indices = [i for i, element in enumerate(self._tab[key]) if element == value]
        if indices:
            if len(indices) == 1:
                return indices[0]
            else:
                return indices
        return None

    def get_row_by_primary_key(self, value):
        row_index = self.get_row_index(self._primary_key, value)
        return {c: self._tab[c][row_index] for c in self.columns}


def to_snake_case(camel_case):
    return re.sub(r"(?<!^)(?=[A-Z])", "_", camel_case).lower()


def parse_sqlalchemy_table(sa_table):
    columns, primary = [], None
    for column in sa_table.__table__.columns:
        columns.append(to_snake_case(column.name))
        if column.primary_key and not primary:
            primary = columns[-1]
    return columns, primary


class MotionCorrectionTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(sqlalchemy.MotionCorrection)
        columns.append("created_time_stamp")
        columns.append("drift_data")
        # columns.append("job_string")
        super().__init__(
            columns,
            prim_key,
            unique="micrograph_full_path",
            counters="image_number",
            required="micrograph_full_path",
        )


class CTFTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(sqlalchemy.CTF)
        super().__init__(
            columns,
            prim_key,
            unique="motion_correction_id",
            required="motion_correction_id",
        )


class ParticlePickerTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(sqlalchemy.ParticlePicker)
        columns.extend(
            [
                "micrograph_full_path",
                "mc_image_full_path",
                "particle_coordinates",
                "job_string",
            ]
        )
        super().__init__(
            columns,
            prim_key,
            unique=["micrograph_full_path", "job_string"],
            required="first_motion_correction_id",
        )


class ParticleClassificationGroupTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(
            sqlalchemy.ParticleClassificationGroup
        )
        columns.append("job_string")
        columns.append("class_images_stack")
        columns.append("class_images_modification_time")
        super().__init__(
            columns,
            prim_key,
            unique="job_string",
            required=["job_string", "particle_picker_id"],
        )


class ParticleClassificationTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(sqlalchemy.ParticleClassification)
        columns.append("job_string")
        super().__init__(
            columns,
            prim_key,
            unique=["job_string", "class_number"],
            required=["class_number", "particle_classification_group_id"],
        )


class CryoemInitialModelTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(sqlalchemy.CryoemInitialModel)
        columns.append("ini_model_job_string")
        columns.append("particle_classification_id")
        super().__init__(
            columns,
            prim_key,
            unique="ini_model_job_string",
            append="particle_classification_id",
            required="particle_classification_id",
        )


class RelativeIceThicknessTable(Table):
    def __init__(self):
        columns, prim_key = parse_sqlalchemy_table(sqlalchemy.RelativeIceThickness)
        super().__init__(
            columns,
            prim_key,
            unique="motion_correction_id",
            required="motion_correction_id",
        )


@functools.singledispatch
def insert(primary_table, end_time, source, relion_options, **kwargs):
    raise ValueError(f"{primary_table!r} is not a known Table")


@insert.register(MotionCorrectionTable)
def _(
    primary_table: MotionCorrectionTable,
    end_time,
    source,
    relion_options,
    row,
):
    row.update(
        {
            "dose_per_frame": relion_options.motioncor_doseperframe,
            "patches_used_x": relion_options.motioncor_patches_x,
            "patches_used_y": relion_options.motioncor_patches_y,
        }
    )
    pid = primary_table.add_row(row)
    return pid


@insert.register(CTFTable)
def _(
    primary_table: CTFTable,
    end_time,
    source,
    relion_options,
    row,
):
    row.update(
        {
            "box_size_x": relion_options.ctffind_boxsize,
            "box_size_y": relion_options.ctffind_boxsize,
            "min_resolution": relion_options.ctffind_minres,
            "max_resolution": relion_options.ctffind_maxres,
            "min_defocus": relion_options.ctffind_defocus_min,
            "max_defocus": relion_options.ctffind_defocus_max,
            "defocus_step_size": relion_options.ctffind_defocus_step,
        }
    )
    pid = primary_table.add_row(row)
    return pid


@insert.register(ParticlePickerTable)
def _(
    primary_table: ParticlePickerTable,
    end_time,
    source,
    relion_options,
    row,
):
    row.update(
        {
            "particle_picking_template": relion_options.cryolo_gmodel,
            "particle_diameter": int(
                relion_options.extract_boxsize
                * relion_options.angpix
                * relion_options.motioncor_binning
            )
            / 10,
        }
    )
    pid = primary_table.add_row(row)
    return pid


@insert.register(ParticleClassificationGroupTable)
def _(
    primary_table: ParticleClassificationGroupTable,
    end_time,
    source,
    relion_options,
    row,
):
    if row.get("type") == "2D":
        ncpb = relion_options.class2d_nr_classes
    elif row.get("type") == "3D":
        ncpb = relion_options.class3d_nr_classes
    else:
        ncpb = 0
    row.update(
        {
            "number_of_particles_per_batch": relion_options.batch_size,
            "number_of_classes_per_batch": ncpb,
            "symmetry": relion_options.symmetry,
        }
    )
    pid = primary_table.add_row(row)
    return pid


@insert.register(ParticleClassificationTable)
def _(
    primary_table: ParticleClassificationTable,
    end_time,
    source,
    relion_options,
    row,
):
    pid = primary_table.add_row(row)
    return pid


@insert.register(CryoemInitialModelTable)
def _(
    primary_table: CryoemInitialModelTable,
    end_time,
    source,
    relion_options,
    row,
):
    row["number_of_particles"] = row["init_model_number_of_particles"][
        row["init_model_class_num"]
    ]
    row.update({"resolution": relion_options.inimodel_resol_final})
    pid = primary_table.add_row(row)
    return pid


@insert.register(RelativeIceThicknessTable)
def _(
    primary_table: RelativeIceThicknessTable,
    end_time,
    source,
    relion_options,
    row,
):
    pid = primary_table.add_row(row)
    return pid



"""
dls_options
-----------

Default options file for eBIC. Uses the DLS Hamilton GPU cluster.

Defaults for all options not defined here are as in the original relion_it.py script.
"""

from __future__ import annotations

#### crYOLO options
# Run Cryolo picking or autopicking
autopick_do_cryolo = True
# Threshold for cryolo autopicking (higher the threshold the more *discriminative the cryolo picker) ((* But beware it may still not be picking what you want! ))
cryolo_threshold = 0.15
# Finetune the cryolo general model by selecting good classes from 2D classification
cryolo_finetune = False

# Location of the cryolo specific files
cryolo_config = "/dls_sw/apps/EM/crYOLO/phosaurus_models/config.json"
cryolo_gmodel = (
    "/dls_sw/apps/EM/crYOLO/phosaurus_models/gmodel_phosnet_202005_N63_c17.h5"
)

# Running options for cryolo
cryolo_pick_gpus = "0 1 2 3"
cryolo_submit_to_queue = True
cryolo_queue_submission_template = "/dls_sw/apps/EM/crYOLO/qsub_cryolo_template_rh7"

do_second_pass = False

#### Other site configuration options
# Assume modules are loaded so these executables can be found by name alone

# Exectutable of UCSF MotionCor2
motioncor_exe = "MotionCor2"

# Executable to Kai Zhang's Gctf
gctf_exe = "Gctf"

# Executable for Alexis Rohou's CTFFIND4
ctffind4_exe = "ctffind"

queue_submission_template = "/dls_sw/apps/EM/relion/qsub_template_hamilton"
queue_submission_template_cpu = "/dls_sw/apps/EM/relion/qsub_template_hamilton_cpu"
queue_submission_template_cpu_smp = (
    "/dls_sw/apps/EM/relion/qsub_template_hamilton_cpu_smp"
)

motioncor_gpu = ""
motioncor_do_own = False
motioncor_mpi = 4
motioncor_threads = 10
motioncor_other_args = "--do_at_most 200"
motioncor_submit_to_queue = True

gctf_gpu = ""
ctffind_mpi = 40
use_ctffind_instead = True
ctffind_submit_to_queue = True

autopick_do_gpu = True
autopick_gpu = ""
autopick_mpi = 4
autopick_submit_to_queue = True

extract_mpi = 40
extract_submit_to_queue = True

batch_size = 50000

refine_do_gpu = True
refine_gpu = ""
refine_mpi = 5
refine_threads = 8
refine_scratch_disk = ""
# Hamilton nodes only have ~100 GB in /tmp, so don't use this
refine_preread_images = True
refine_submit_to_queue = True

minimum_batch_size = 250


#### Default options for the GUI

# (These might belong better in a different template file but they'll work here for now)

voltage = 300
Cs = 2.7
import_images = "Movies/*.tiff"



from __future__ import annotations

import logging
import time
from pathlib import Path

import mrcfile
import numpy as np
import PIL.Image
from PIL import ImageDraw, ImageEnhance, ImageFilter

logger = logging.getLogger("relion.zocalo.images_service_plugin")


def mrc_to_jpeg(plugin_params):
    filename = plugin_params.parameters("file")
    allframes = plugin_params.parameters("all_frames")
    if not filename or filename == "None":
        logger.error("Skipping mrc to jpeg conversion: filename not specified")
        return False
    filepath = Path(filename)
    if not filepath.is_file():
        logger.error(f"File {filepath} not found")
        return False
    start = time.perf_counter()
    try:
        with mrcfile.open(filepath) as mrc:
            data = mrc.data
    except ValueError:
        logger.error(
            f"File {filepath} could not be opened. It may be corrupted or not in mrc format"
        )
        return False
    outfile = filepath.with_suffix(".jpeg")
    outfiles = []
    if len(data.shape) == 2:
        mean = np.mean(data)
        sdev = np.std(data)
        sigma_min = mean - 3 * sdev
        sigma_max = mean + 3 * sdev
        data = np.ndarray.copy(data)
        data[data < sigma_min] = sigma_min
        data[data > sigma_max] = sigma_max
        data = data - data.min()
        data = data * 255 / data.max()
        data = data.astype("uint8")
        im = PIL.Image.fromarray(data, mode="L")
        try:
            im.save(outfile)
        except FileNotFoundError:
            logger.error(
                f"Trying to save to file {outfile} but directory does not exist"
            )
            return False
    elif len(data.shape) == 3:
        if allframes:
            for i, frame in enumerate(data):
                frame = frame - frame.min()
                frame = frame * 255 / frame.max()
                frame = frame.astype("uint8")
                im = PIL.Image.fromarray(frame, mode="L")
                frame_outfile = str(outfile).replace(".jpeg", f"_{i+1}.jpeg")
                try:
                    im.save(frame_outfile)
                except FileNotFoundError:
                    logger.error(
                        f"Trying to save to file {frame_outfile} but directory does not exist"
                    )
                    return False
                outfiles.append(frame_outfile)
        else:
            data = data - data[0].min()
            data = data * 255 / data[0].max()
            data = data.astype("uint8")
            im = PIL.Image.fromarray(data[0], mode="L")
            try:
                im.save(outfile)
            except FileNotFoundError:
                logger.error(
                    f"Trying to save to file {outfile} but directory does not exist"
                )
                return False
    timing = time.perf_counter() - start

    logger.info(
        f"Converted mrc to jpeg {filename} -> {outfile} in {timing:.1f} seconds",
        extra={"image-processing-time": timing},
    )
    if outfiles:
        return outfiles
    return outfile


def picked_particles(plugin_params):
    basefilename = plugin_params.parameters("file")
    if basefilename.endswith(".jpeg"):
        logger.info(f"Replacing jpeg extension with mrc extension for {basefilename}")
        basefilename = basefilename.replace(".jpeg", ".mrc")
    coords = plugin_params.parameters("coordinates")
    if not coords:
        logger.warning(f"No coordinates provided for {basefilename}")
        # If there were no coordinates don't bother nacking the message
        return True
    angpix = plugin_params.parameters("angpix")
    diam = plugin_params.parameters("diameter")
    contrast_factor = plugin_params.parameters("contrast_factor", default=6)
    outfile = plugin_params.parameters("outfile")
    if not outfile:
        logger.error(f"Outfile incorrectly specified: {outfile}")
        return False
    if not Path(basefilename).is_file():
        logger.error(f"File {basefilename} not found")
        return False
    radius = (diam / angpix) // 2
    start = time.perf_counter()
    try:
        with mrcfile.open(basefilename) as mrc:
            data = mrc.data
    except ValueError:
        logger.error(
            f"File {basefilename} could not be opened. It may be corrupted or not in mrc format"
        )
        return False
    except FileNotFoundError:
        logger.error(f"File {basefilename} could not be opened")
        return False
    mean = np.mean(data)
    sdev = np.std(data)
    sigma_min = mean - 3 * sdev
    sigma_max = mean + 3 * sdev
    data = np.ndarray.copy(data)
    data[data < sigma_min] = sigma_min
    data[data > sigma_max] = sigma_max
    data = data - data.min()
    data = data * 255 / data.max()
    data = data.astype("uint8")
    with PIL.Image.fromarray(data).convert(mode="RGB") as bim:
        enhancer = ImageEnhance.Contrast(bim)
        enhanced = enhancer.enhance(contrast_factor)
        fim = enhanced.filter(ImageFilter.BLUR)
        dim = ImageDraw.Draw(fim)
        for x, y in coords:
            dim.ellipse(
                [
                    (float(x) - radius, float(y) - radius),
                    (float(x) + radius, float(y) + radius),
                ],
                width=8,
                outline="#f58a07",
            )
        try:
            fim.save(outfile)
        except FileNotFoundError:
            logger.error(
                f"Trying to save to file {outfile} but directory does not exist"
            )
            return False
    timing = time.perf_counter() - start
    logger.info(
        f"Particle picker image {outfile} saved in {timing:.1f} seconds",
        extra={"image-processing-time": timing},
    )
    return outfile


def mrc_central_slice(plugin_params):
    filename = plugin_params.parameters("file")
    if not filename or filename == "None":
        logger.error("Skipping mrc to jpeg conversion: filename not specified")
        return False
    filepath = Path(filename)
    if not filepath.is_file():
        logger.error(f"File {filepath} not found")
        return False
    start = time.perf_counter()
    try:
        with mrcfile.open(filepath) as mrc:
            data = mrc.data
    except ValueError:
        logger.error(
            f"File {filepath} could not be opened. It may be corrupted or not in mrc format"
        )
        return False
    outfile = str(filepath.with_suffix("")) + "_thumbnail.jpeg"
    if len(data.shape) != 3:
        logger.error(
            f"File {filepath} is not 3-dimensional. Cannot extract central slice"
        )
        return False

    # Extract central slice
    total_slices = data.shape[0]
    central_slice_index = int(total_slices / 2)
    central_slice_data = data[central_slice_index, :, :]

    # Write as jpeg
    central_slice_data = central_slice_data - central_slice_data[0].min()
    central_slice_data = central_slice_data * 255 / central_slice_data[0].max()
    central_slice_data = central_slice_data.astype("uint8")
    im = PIL.Image.fromarray(central_slice_data, mode="L")
    im.thumbnail((512, 512))
    try:
        im.save(outfile)
    except FileNotFoundError:
        logger.error(f"Trying to save to file {outfile} but directory does not exist")
        return False
    timing = time.perf_counter() - start

    logger.info(
        f"Converted mrc to jpeg {filename} -> {outfile} in {timing:.1f} seconds",
        extra={"image-processing-time": timing},
    )

    Path(outfile).chmod(0o740)
    return outfile


def mrc_to_apng(plugin_params):
    filename = plugin_params.parameters("file")

    if not filename or filename == "None":
        logger.error("Skipping mrc to jpeg conversion: filename not specified")
        return False
    filepath = Path(filename)

    filepath.chmod(0o740)

    if not filepath.is_file():
        logger.error(f"File {filepath} not found")
        return False
    start = time.perf_counter()
    try:
        with mrcfile.open(filepath) as mrc:
            data = mrc.data
    except ValueError:
        logger.error(
            f"File {filepath} could not be opened. It may be corrupted or not in mrc format"
        )
        return False
    outfile = str(filepath.with_suffix("")) + "_movie.png"

    if len(data.shape) == 3:
        images_to_append = []
        for i, frame in enumerate(data):
            frame = frame - frame.min()
            frame = frame * 255 / frame.max()
            frame = frame.astype("uint8")
            im = PIL.Image.fromarray(frame, mode="L")
            im.thumbnail((512, 512))
            images_to_append.append(im)
        try:
            im.save(outfile, save_all=True, append_images=images_to_append)
        except FileNotFoundError:
            logger.error(
                f"Trying to save to file {outfile} but directory does not exist"
            )
            return False
    else:
        logger.error(f"File {filepath} is not a 3D volume")
    timing = time.perf_counter() - start
    logger.info(
        f"Converted mrc to apng {filename} -> {outfile} in {timing:.1f} seconds"
    )

    Path(outfile).chmod(0o740)
    return outfile
